#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.1.0"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2019 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from functools import partial
import re
from objutils.checksums import lrc, COMPLEMENT_ONES
from objutils.utils import make_list
import objutils.hexfile as hexfile
import objutils.utils as utils

S0  = 1
S1  = 2
S2  = 3
S3  = 4
S5  = 5
S7  = 6
S8  = 7
S9  = 8
SYM = 9

BIAS = {
    S0 : 3,
    S1 : 3,
    S2 : 4,
    S3 : 5,
    S5 : 2,
    S7 : 5,
    S8 : 4,
    S9 : 3
}

SYMBOLTABLE = re.compile(r"(^\$\$\s+(?P<modulename>\S*)(?P<symbols>.*?)\$\$)",re.MULTILINE|re.DOTALL)
SYMBOL = re.compile(r'\s+(?P<symbol>.*?)\s+\$(?P<value>.+)',re.MULTILINE|re.DOTALL)

class Reader(hexfile.Reader):
    FORMAT_SPEC = (
        (S0, "S0LLAAAADDCC"),
        (S1, "S1LLAAAADDCC"),
        (S2, "S2LLAAAAAADDCC"),
        (S3, "S3LLAAAAAAAADDCC"),
        (S5, "S5LLAAAACC"),
        (S7, "S7LLAAAAAAAACC"),
        (S8, "S8LLAAAAAACC"),
        (S9, "S9LLAAAACC"),
    )

    def load(self, fp, **kws):
        if isinstance(fp, str):
            fp = open(fp, "rb")
        data = self.read(fp)


        ## todo: extract Symbols and wipe them out.
        """
        symbol_tables = SYMBOLTABLE.findall(data)
        if symbol_tables:
            self._strip_symbols(symbol_tables)
        records = SYMBOLTABLE.sub('', data).strip()
        """

        return data

    def check_line(self, line, format_type):
        # todo: Fkt.!!!
        if format_type in (S0, S1, S5, S9):
            checksum_of_address = ((line.address & 0xff00) >> 8) + (line.address & 0xff)
        elif format_type in (S2, S8):
            checksum_of_address = ((line.address & 0xff0000) >> 16) + ((line.address & 0xff00) >> 8) + (line.address & 0xff)
        elif format_type in (S3, S7):
            checksum_of_address = (((line.address & 0xff000000) >> 24)+((line.address & 0xff0000) >> 16) +
                ((line.address & 0xff00) >>8 )+(line.address & 0xff)
            )
        else:
            raise TypeError("Invalid format type '{0!s}'.".format(format_type))
        if hasattr(line, 'chunk'):
            checksum = (~(sum([line.length,checksum_of_address]) + sum(line.chunk))) & 0xff
        else:
            checksum = (~(sum([line.length,checksum_of_address]))) & 0xff
        if line.checksum != checksum:
            raise hexfile.InvalidRecordChecksumError()
        line.length -= BIAS[format_type]   # calculate actual data length.
        if hasattr(line, 'chunk') and line.length and (line.length != len(line.chunk)):
            raise hexfile.InvalidRecordLengthError("Byte count doesn't match length of actual data.")

    def is_data_line(self, line, format_type):
        return format_type in (S1, S2, S3)

    def special_processing(self, line, format_type):
        if format_type == S0:
            #print("S0: [{}]".format(line.chunk))
            pass
        elif format_type == S5:
            #print "S5: [%s]" % line.chunk
            start_address = line.address
        elif format_type == S7:
            start_address = line.address
            #print "Startaddress[S7]: %u" % start_address
            #print "32-Bit Start-Address: ", hex(start_address)
        elif format_type == S8:
            start_address = line.address
            #print "Startaddress[S8]: %u" % start_address
            #print "24-Bit Start-Address: ", hex(start_address)
        elif format_type == S9:
            start_address = line.address
            #print "Startaddress[S9]: %u" % start_address
            #print "16-Bit Start-Address: ", hex(start_address)

    def _strip_symbols(self, symbol_tables):
        self.symbols=[]
        for _, module_name, symbol_table in symbol_tables:
            sb = []
            for symbol in symbol_table.splitlines():
                ma = SYMBOL.match(symbol)
                if ma:
                    #print ma.groupdict()
                    gd = ma.groupdict()
                    sb.append((gd['symbol'], int(gd['value'], 16)))
            self.symbols.append(sb)
        #print self.symbols


class Writer(hexfile.Writer):
    record_type = None
    s5record = False
    start_address = None

    MAX_ADDRESS_BITS = 32

    checksum = partial(lrc, width = 8, comp = COMPLEMENT_ONES)

    def pre_processing(self, image):
        if self.record_type is None:
            if hasattr(image, "sections"):
                last_segment = sorted(image.sections, key = lambda s: s.start_address)[-1]
            else:
                last_segment = image
            highest_address = last_segment.start_address + last_segment.length
            if highest_address <= 0x000000ffff:
                self.record_type = 1
            elif highest_address <= 0x00ffffff:
                self.record_type = 2
            elif highest_address <= 0xffffffff:
                self.record_type = 3
        self.address_mask = "%0{0:d}X".format((self.record_type + 1) * 2 )
        self.offset = self.record_type + 2


    def srecord(self, record_type, length, address, data = None):
        if data is None:
            data = []
        length += self.offset
        address_bytes = utils.int_to_array(address)
        checksum = self.checksum(make_list(address_bytes, length, data))
        mask = "S%u%02X{0!s}%s%02X".format(self.address_mask)
        return mask % (record_type, length, address, Writer.hex_bytes(data), checksum)

    def compose_row(self, address, length, row):
        self.record_count += 1
        return self.srecord(self.record_type, length, address, row)

    def compose_header(self, meta):
        self.record_count = 0
        result = []
        if S0 in meta:  # Usually only one S0 record, but be tolerant.
            for meta in meta[S0]:
                result.append(self.srecord(0, len(meta.chunk), meta.address, meta.chunk))
        return '\n'.join(result)

    def compose_footer(self, meta):
        result = []
        if self.s5record:
            result.append(self.srecord(5, 0, self.record_count))
        if self.start_address is not None:
            if self.record_type == 1:    # 16bit.
                if S9 in meta:
                    s9 = meta[S9][0]
                    result.append(self.srecord(9, 0, s9.address))
                else:
                    result.append(self.srecord(9, 0, self.start_address))
            elif self.record_type == 2:  # 24bit.
                if S8 in meta:
                    s8 = meta[S8][0]
                    result.append(self.srecord(8, 0, s8.address))
                else:
                    result.append(self.srecord(8, 0, self.start_address))
            elif self.record_type == 3:  # 32bit.
                if S7 in meta:
                    s7 = meta[S7][0]
                    result.append(self.srecord(7, 0, s7.address))
                else:
                    result.append(self.srecord(7, 0, self.start_address))
        return '\n'.join(result)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import division

__version__ = "0.1.0"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2019 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
from functools import partial
import os
import re
import objutils.hexfile as hexfile
from objutils.utils import create_string_buffer, slicer, PYTHON_VERSION
from objutils import checksums
import objutils.utils as utils


DATA_ABS    = 1
DATA_INC    = 2
DATA_REL    = 3
EOF         = 4

PREFIX      = '$'

MAPPING = dict(enumerate(chr(n) for n in range(37, 123) if not n in (42, )))
REV_MAPPING = {ord(value): key for key, value in MAPPING.items()}
NULLS = re.compile(r'\0*\s*!M\s*(.*)', re.DOTALL | re.M)
VALID_CHARS = re.compile("^\{0}[{1}]+$".format(PREFIX, re.escape(''.join(MAPPING.values()))))

atoi16 = partial(int, base = 16)


class Reader(hexfile.Reader):

    FORMAT_SPEC = (
        (DATA_ABS,  "CCLL0000AAAAAAAADD"),
        (DATA_INC,  "CCLL0001DD"),
        (DATA_REL,  "CCLL0002AAAAAAAADD"),
        (EOF,       "00000000")
    )

    def decode(self, fp):
        self.last_address = 0    # TODO: decode!
        out_lines = []
        for line in fp.readlines():
            line = line.strip()
            startSym, line = line[0], line[1:]

            if startSym != PREFIX:
                pass # todo: FormatError!!!

            if (len(line) % 5) != 0:
                pass # todo: FormatError!!!

            values = []
            for quintuple in self.split_quintuples(line):
                value = self.convert_quintuple(quintuple)
                values.append("{0:08X}".format(value))
            out_lines.append(''.join(values))
        return '\n'.join(out_lines)

    def read(self, fp):
        return super(Reader, self).read(
            create_string_buffer(
                bytearray(self.decode(fp), "ascii")
            )
        )

    def convert_quintuple(self, quintuple):
        res = 0         # reduce(lambda accu, x: (accu * 85) + x, value, 0)
        for ch  in quintuple:
            v = REV_MAPPING[bytearray((ch, ))[0]]
            res = v + (res * 85)
        return res

    def split_quintuples(self, line):
        res = []
        for i in range(0, len(line), 5):
            res.append(line[i : i + 5])
        return res

    def check_line(self, line, format_type):
        if format_type == EOF:
            return True
        line.length -= 4
        if line.length != len(line.chunk):
            line.chunk = line.chunk[ : line.length] # Cut padding.
        if format_type == DATA_ABS:
            tmp = 0
            self.last_address = line.address + line.length
        elif format_type == DATA_INC:
            tmp = 1
            line.address = self.last_address
        elif format_type == DATA_REL:
            self.error("relative adressing not supported.")
            tmp = 2
        else:
            self.error("Invalid format type: '{0}'".format(format_type))
            tmp = 0
        checksum = checksums.lrc(utils.make_list(tmp, line.length + 4, utils.int_to_array(line.address), line.chunk), 8, checksums.COMPLEMENT_TWOS)
        if line.checksum != checksum:
            raise hexfile.InvalidRecordChecksumError()

    def is_data_line(self, line, format_type):
        return format_type in (DATA_ABS, DATA_INC, DATA_REL)

    def probe(self, fp):
        for idx, line in enumerate(fp, 1):
            if not VALID_CHARS.match(line.decode()):
                fp.seek(0, os.SEEK_SET)
                return False
            if idx > 3:
                break
        fp.seek(0, os.SEEK_SET)
        return super(Reader, self).probe(
            create_string_buffer(bytearray(self.decode(fp), "ascii"))
        )


class Writer(hexfile.Writer):

    MAX_ADDRESS_BITS = 16

    def compose_row(self, address, length, row):
        tmp = 0 # TODO: format type!?
        checksum = checksums.lrc(utils.make_list(tmp, length + 4, utils.int_to_array(address), row), 8, checksums.COMPLEMENT_TWOS)
        if length < self.row_length:
            lengthToPad = self.row_length - length
            padding = [0] * (lengthToPad)
            row.extend(padding)
        line = "{0:02X}{1}0000{2:08X}{3}".format(checksum, length - 2, address, Writer.hex_bytes(row))
        return line

    def compose_footer(self, meta):
        return "00000000"

    def post_processing(self, data):
        result = []
        for line in data.splitlines():
            if len(line) % 4:
                self.error("Size of line must be a multiple of 4.")
                continue
            res = []
            for item in slicer(line, 8, atoi16):
                item = self.convert_quintuple(item)
                res.append(item)
            result.append("{0}{1}".format(PREFIX, ''.join(res)))
        if PYTHON_VERSION.major == 3:
            return bytes('\n'.join(result), "ascii")
        else:
            return bytes('\n'.join(result))

    def convert_quintuple(self, value):
        result = []
        while value:
            result.append(MAPPING[value % 85])
            value //= 85
        if len(result) < 5:
            result.extend([MAPPING[0]] * (5 - len(result)))
        return ''.join(reversed(result))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.1.0"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2019 by Christoph Schueler <github.com/Christoph2,
                                        cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from functools import partial
import operator
import objutils.hexfile as hexfile
from objutils.checksums import lrc, COMPLEMENT_TWOS
import objutils.utils as utils
import objutils.checksums as checksums

DATA                        = 0
EOF                         = 1
EXTENDED_SEGMENT_ADDRESS    = 2
START_SEGMENT_ADDRESS       = 3
EXTENDED_LINEAR_ADDRESS     = 4
START_LINEAR_ADDRESS        = 5


class Reader(hexfile.Reader):

    FORMAT_SPEC = (
        (hexfile.TYPE_FROM_RECORD, ":LLAAAATTDDCC"),
        )

    def __init__(self):
        super(Reader,self).__init__()
        self.segmentAddress = 0

    def check_line(self, line, format_type):
        if line.length != len(line.chunk):
            raise hexfile.InvalidRecordLengthError("Byte count doesn't match length of actual data.")
        checksum = checksums.lrc(utils.make_list(line.type, line.length, utils.int_to_array(line.address), line.chunk), 8, checksums.COMPLEMENT_TWOS)
        if line.checksum != checksum:
            raise hexfile.InvalidRecordChecksumError()

    def is_data_line(self, line, format_type):
        if line.type == DATA:
            return True
        else:
            return False

    def calculate_extended_address(self, line, shift_by, name):
            if len(line.chunk) == 2:
                segment = ((line.chunk[0]) << 8) | (line.chunk[1])
                line.add_processing_instruction(('segment', segment))
                self._address_calculator = partial(operator.add, segment << shift_by)
                self.debug("EXTENDED_{0}_ADDRESS: {1:#X}".format(name.upper(), segment))
            else:
                self.error("Bad Extended {0} Address at line #{1}.".format(name, line.line_number))

    def special_processing(self, line, format_type):
        if line.type == DATA:
            line.address = self._address_calculator(line.address)
        elif line.type == EXTENDED_SEGMENT_ADDRESS:
            self.calculate_extended_address(line, 4, "Segment")
        elif line.type == START_SEGMENT_ADDRESS:
            if len(line.chunk) == 4:
                cs = ((line.chunk[0]) << 8) | (line.chunk[1])
                ip = ((line.chunk[2]) << 8) | (line.chunk[3])
                line.add_processing_instruction(('cs', cs))
                line.add_processing_instruction(('ip', ip))
                self.debug("START_SEGMENT_ADDRESS: {0}:{1}".format(hex(cs), hex(ip)))
            else:
                self.error("Bad Segment Address at line %{0:u}.".format(line.line_number))
        elif line.type == EXTENDED_LINEAR_ADDRESS:
            self.calculate_extended_address(line, 16, "Linear")
        elif line.type == START_LINEAR_ADDRESS:
            if len(line.chunk) == 4:
                eip = ((line.chunk[0]) << 24) | ((line.chunk[1]) << 16) | ((line.chunk[2]) << 8) | (line.chunk[3])
                line.add_processing_instruction(('eip', eip))
                self.debug("START_LINEAR_ADDRESS: {0}".format(hex(eip)))
            else:
                self.error("Bad Linear Address at line #{0:d}.".format(line.line_number))
        elif line.type == EOF:
            pass
        else:
            self.warn("Invalid record type [{0:u}] at line {1:u}".format(line.type, line.line_number))

    _address_calculator = utils.identity


def divmod(a, b):
    return a / b, a % b

class Writer(hexfile.Writer):

    MAX_ADDRESS_BITS = 32
    checksum = partial(lrc, width = 8, comp = COMPLEMENT_TWOS)

    def pre_processing(self, image):
        self.previosAddress = None
        self.start_address = 0

    def compose_row(self, address, length, row):
        result = ''
        seg, offs = divmod(address, 0x10000)
        h, l = self.word_to_bytes(address)
        if offs != self.previosAddress:
            #print("NEQ: {0:04x} [{1:04x}:{2:04x}]".format(address, seg, offs))
            if address > 0xffff:
                if address > 0xfffff:
                    segHi, segLo = self.word_to_bytes(seg)
                    result = ":02000004{0:04X}{1:02X}\n".format(int(seg), self.checksum(list((2, 4, segHi, segLo))))
                else:
                    seg = int(seg) << 12
                    segHi, segLo = self.word_to_bytes(seg)
                    result = ":02000002{0:04X}{1:02X}\n".format(int(seg), self.checksum(list((2, 2, segHi, segLo))))
        address = offs
        checksum = self.checksum(list((length, h, l)) + list(row))
        self.previosAddress = offs + length
        result += ":{0:02X}{1:04X}{2:02X}{3!s}{4:02X}".format(length, address, DATA, Writer.hex_bytes(row), checksum)
        return result

    def compose_footer(self, meta):
        h, l = self.word_to_bytes((self.start_address))
        return ":00{1:04X}{0:02X}{2:02X}\n".format(EOF, self.start_address, self.checksum(list((h, l, EOF))))




import array
import math
import struct

import pytest

from objutils.section import (Section, filler, INT8_RANGE, INT16_RANGE,
        INT32_RANGE, INT64_RANGE, UINT8_RANGE, UINT16_RANGE, UINT32_RANGE,
        UINT64_RANGE)

from objutils.exceptions import InvalidAddressError

try:
    import numpy as np
except ImportError:
    NUMPY_SUPPORT = False
else:
    NUMPY_SUPPORT = True

def test_default_section():
    section = Section()
    #section = Section2(data=[10, 20, 30, 40])
    assert section.start_address == 0
    assert section.length == 0
    assert section.data == b''

def test_default_startAddress():
    section = Section(start_address = 0x10000)
    assert section.start_address == 0x10000
    assert section.length == 0
    assert section.data == b''

def test_initialize_from_list():
    section = Section(data = [10, 20, 30, 40])
    assert section.start_address == 0
    assert section.length == 4
    assert section.data == b'\n\x14\x1e('

def test_initialize_from_array1():
    section = Section(data = array.array('B', [10, 20, 30, 40]))
    assert section.start_address == 0
    assert section.length == 4
    assert section.data == b'\n\x14\x1e('

def test_initialize_from_array2():
    section = Section(data = array.array('H', [0x1010, 0x2020, 0x3030, 0x4040]))
    assert section.start_address == 0
    assert section.length == 8
    assert section.data == b'\x10\x10  00@@'

def test_initialize_from_range():
    section = Section(data = range(16))
    assert section.start_address == 0
    assert section.length == 16
    assert section.data == b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f'

def test_initialize_from_string():
    section = Section(data = "abcd")
    assert section.start_address == 0
    assert section.length == 4
    assert section.data == b'abcd'

def test_adress_in_range1():
    s0 = Section(start_address = 0x10, data = "hello")
    assert 0x10 in s0

def test_adress_in_range2():
    s0 = Section(start_address = 0x10, data = "hello")
    assert 0x14 in s0

def test_adress_out_of_range1():
    s0 = Section(start_address = 0x10, data = "hello")
    assert not 0x09 in s0

def test_adress_out_of_range2():
    s0 = Section(start_address = 0x10, data = "hello")
    assert not 0x15 in s0

def test_copy_data_from_other_section():
    data = Section(data = "abcd", start_address = 0x8000)
    section = Section(data = data)
    assert section.start_address == 0
    assert section.length == 4
    assert section.data == b'abcd'

def test_copy_really_works():
    section1 = Section(data = range(16), start_address = 0x8000)
    section2 = Section(data = section1)
    assert section2.start_address == 0
    assert section2.length == 16
    assert section2.data == b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f'
    assert section2.data == section1.data
    section2.write_numeric(0, 0xff, "uint8_le")
    assert section2.data == b'\xff\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f'
    assert section1.data != section2.data

def test_single_int_not_permitted():
    with pytest.raises(ValueError):
        Section(data = 42)

def test_filler1():
    assert filler(0x00, 16) == b'\x00' * 16

def test_filler2():
    assert filler(0xff, 16) == b'\xff' * 16

def test_filler_fails1():
    with pytest.raises(ValueError):
        filler(0xff, 0)

def test_filler_fails2():
    with pytest.raises(ValueError):
        filler(1000, 10)

def test_filler_fails3():
    with pytest.raises(TypeError):
        filler("hello", 10)

def test_filler_fails4():
    with pytest.raises(TypeError):
        filler(0x00, "hello")

def test_int8_range():
    assert INT8_RANGE == (-128, 127)

def test_uint8_range():
    assert UINT8_RANGE == (0, 255)

def test_uint16_range():
    assert UINT16_RANGE == (0, 65535)

def test_int16_range():
    assert INT16_RANGE == (-32768, 32767)

def test_int32_range():
    assert INT32_RANGE == (-2147483648, 2147483647)

def test_uint32_range():
    assert UINT32_RANGE == (0, 4294967295)

def test_int64_range():
    assert INT64_RANGE == (-9223372036854775808, 9223372036854775808)

def test_uint64_range():
    assert UINT64_RANGE == (0, 18446744073709551615)

@pytest.fixture
def filler_0_16():
    return Section(data = filler(0x00, 16))

def test_write_uint8_le1(filler_0_16):
    filler_0_16.write_numeric(0, 0xff, "uint8_le")
    assert filler_0_16.data == bytearray([0xff, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

def test_write_uint8_le2(filler_0_16):
    with pytest.raises(struct.error):
        filler_0_16.write_numeric(0, UINT8_RANGE.lower - 1, "uint8_le")

def test_write_uint8_le3(filler_0_16):
    with pytest.raises(struct.error):
        filler_0_16.write_numeric(0, UINT8_RANGE.upper + 1, "uint8_le")

def test_write_uint8_be1(filler_0_16):
    filler_0_16.write_numeric(0, 0xff, "uint8_be")
    assert filler_0_16.data == bytearray([0xff, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

def test_write_uint8_be2(filler_0_16):
    with pytest.raises(struct.error):
        filler_0_16.write_numeric(0, UINT8_RANGE.lower - 1, "uint8_be")

def test_write_uint8_be3(filler_0_16):
    with pytest.raises(struct.error):
        filler_0_16.write_numeric(0, UINT8_RANGE.upper + 1, "uint8_be")

def test_write_int8_le1(filler_0_16):
    filler_0_16.write_numeric(0, -100, "int8_le")
    assert filler_0_16.data == bytearray([0x9c, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

def test_write_int8_le2(filler_0_16):
    with pytest.raises(struct.error):
        filler_0_16.write_numeric(0, INT8_RANGE.lower - 1, "int8_le")

def test_write_int8_le3(filler_0_16):
    with pytest.raises(struct.error):
        filler_0_16.write_numeric(0, INT8_RANGE.upper + 1, "int8_le")

def test_write_int8_be1(filler_0_16):
    filler_0_16.write_numeric(0, -100, "int8_be")
    assert filler_0_16.data == bytearray([0x9c, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

def test_write_int8_be2(filler_0_16):
    with pytest.raises(struct.error):
        filler_0_16.write_numeric(0, INT8_RANGE.lower - 1, "int8_be")

def test_write_int8_be3(filler_0_16):
    with pytest.raises(struct.error):
        filler_0_16.write_numeric(0, INT8_RANGE.upper + 1, "int8_be")

##
##
def test_write_byte0(filler_0_16):
    filler_0_16.write_numeric(0, 0xff, "byte")
    assert filler_0_16.data == bytearray([0xff, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

def test_write_byte1(filler_0_16):
    with pytest.raises(struct.error):
        filler_0_16.write_numeric(0, UINT8_RANGE.lower - 1, "byte")

def test_write_byte2(filler_0_16):
    with pytest.raises(struct.error):
        filler_0_16.write_numeric(0, UINT8_RANGE.upper + 1, "byte")

def test_write_byte3(filler_0_16):
    filler_0_16.write_numeric(0, 0xff, "byte")
    assert filler_0_16.data == bytearray([0xff, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

def test_write_byte4(filler_0_16):
    with pytest.raises(struct.error):
        filler_0_16.write_numeric(0, UINT8_RANGE.lower - 1, "byte")

def test_write_byte5(filler_0_16):
    with pytest.raises(struct.error):
        filler_0_16.write_numeric(0, UINT8_RANGE.upper + 1, "byte")

@pytest.mark.skip
def test_write_byte6(filler_0_16):
    filler_0_16.write_numeric(0, -100, "byte")
    assert filler_0_16.data == bytearray([0x9c, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

def test_write_byte7(filler_0_16):
    with pytest.raises(struct.error):
        filler_0_16.write_numeric(0, INT8_RANGE.lower - 1, "byte")

@pytest.mark.skip
def test_write_byte8(filler_0_16):
    with pytest.raises(struct.error):
        filler_0_16.write_numeric(0, INT8_RANGE.upper + 1, "byte")

@pytest.mark.skip
def test_write_byte9(filler_0_16):
    filler_0_16.write_numeric(0, -100, "byte")
    assert filler_0_16.data == bytearray([0x9c, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

def test_write_byte10(filler_0_16):
    with pytest.raises(struct.error):
        filler_0_16.write_numeric(0, INT8_RANGE.lower - 1, "byte")

@pytest.mark.skip
def test_write_byte11(filler_0_16):
    with pytest.raises(struct.error):
        filler_0_16.write_numeric(0, INT8_RANGE.upper + 1, "byte")

##
##

def test_write_uint16_le1(filler_0_16):
    filler_0_16.write_numeric(0, 0x1122, "uint16_le")
    assert filler_0_16.data == bytearray([0x22, 0x11, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

def test_write_uint16_be1(filler_0_16):
    filler_0_16.write_numeric(0, 0x1122, "uint16_be")
    assert filler_0_16.data == bytearray([0x11, 0x22, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

def test_write_int16_le1(filler_0_16):
    filler_0_16.write_numeric(0, -20000, "int16_le")
    assert filler_0_16.data == bytearray([0xe0, 0xb1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

def test_write_int16_be1(filler_0_16):
    filler_0_16.write_numeric(0, -20000, "int16_be")
    assert filler_0_16.data == bytearray([0xb1, 0xe0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

def test_write_uint32_le1(filler_0_16):
    filler_0_16.write_numeric(0, 0x11223344, "uint32_le")
    assert filler_0_16.data == bytearray([0x44, 0x33, 0x22, 0x11, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

def test_write_uint32_be1(filler_0_16):
    filler_0_16.write_numeric(0, 0x11223344, "uint32_be")
    assert filler_0_16.data == bytearray([0x11, 0x22, 0x33, 0x44, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

def test_write_int32_le1(filler_0_16):
    filler_0_16.write_numeric(0, -3000000, "int32_le")
    assert filler_0_16.data == bytearray([0x40, 0x39, 0xd2, 0xff, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

def test_write_int32_be1(filler_0_16):
    filler_0_16.write_numeric(0, -3000000, "int32_be")
    assert filler_0_16.data == bytearray([0xff, 0xd2, 0x39, 0x40, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

def test_write_uint64_le1(filler_0_16):
    filler_0_16.write_numeric(0, 0x1122334455667788, "uint64_le")
    assert filler_0_16.data == bytearray([0x88, 0x77, 0x66, 0x55, 0x44, 0x33, 0x22, 0x11, 0, 0, 0, 0, 0, 0, 0, 0])

def test_write_uint64_be1(filler_0_16):
    filler_0_16.write_numeric(0, 0x1122334455667788, "uint64_be")
    assert filler_0_16.data == bytearray([0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88, 0, 0, 0, 0, 0, 0, 0, 0])

def test_write_int64_le1(filler_0_16):
    filler_0_16.write_numeric(0, -400000000, "int64_le")
    assert filler_0_16.data == bytearray([0x0, 0x7c, 0x28, 0xe8, 0xff, 0xff, 0xff, 0xff, 0, 0, 0, 0, 0, 0, 0, 0])

def test_write_int64_be1(filler_0_16):
    filler_0_16.write_numeric(0, -400000000, "int64_be")
    assert filler_0_16.data == bytearray([0xff, 0xff, 0xff, 0xff, 0xe8, 0x28, 0x7c, 0, 0, 0, 0, 0, 0, 0, 0, 0])

def test_write_float32_le1(filler_0_16):
    filler_0_16.write_numeric(0, math.pi, "float32_le")
    assert filler_0_16.data == bytearray([0xdb, 0x0f, 0x49, 0x40, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

def test_write_float32_be1(filler_0_16):
    filler_0_16.write_numeric(0, math.pi, "float32_be")
    assert filler_0_16.data == bytearray([0x40, 0x49, 0x0f, 0xdb, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

def test_write_float64_le1(filler_0_16):
    filler_0_16.write_numeric(0, math.pi, "float64_le")
    assert filler_0_16.data == bytearray([0x18, 0x2d, 0x44, 0x54, 0xfb, 0x21, 0x09, 0x40, 0, 0, 0, 0, 0, 0, 0, 0])

def test_write_float64_be1(filler_0_16):
    filler_0_16.write_numeric(0, math.pi, "float64_be")
    assert filler_0_16.data == bytearray([0x40, 0x09, 0x21, 0xfb, 0x54, 0x44, 0x2d, 0x18, 0, 0, 0, 0, 0, 0, 0, 0])

def test_write_uint8_negative_offset(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.write_numeric(-1, 0xff, "uint8_le")

def test_write_uint16_negative_offset(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.write_numeric(-1, 0xff, "uint16_le")

def test_write_uint32_negative_offset(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.write_numeric(-1, 0xff, "uint32_le")

def test_write_uint64_negative_offset(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.write_numeric(-1, 0xff, "uint64_le")

def test_write_int8_negative_offset(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.write_numeric(-1, 0xff, "int8_le")

def test_write_int16_negative_offset(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.write_numeric(-1, 0xff, "int16_le")

def test_write_int32_negative_offset(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.write_numeric(-1, 0xff, "int32_le")

def test_write_int64_negative_offset(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.write_numeric(-1, 0xff, "int64_le")

def test_write_float32_negative_offset(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.write_numeric(-1, 3.14159, "float32_le")

def test_write_float64_negative_offset(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.write_numeric(-1, 3.14159, "float64_le")

def test_read_uint8_negative_offset(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.read_numeric(-1, "uint8_le")

def test_read_uint16_negative_offset(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.read_numeric(-1, "uint16_le")

def test_read_uint32_negative_offset(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.read_numeric(-1, "uint32_le")

def test_read_uint64_negative_offset(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.read_numeric(-1, "uint64_le")

def test_read_int8_negative_offset(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.read_numeric(-1, "int8_le")

def test_read_int16_negative_offset(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.read_numeric(-1, "int16_le")

def test_read_int32_negative_offset(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.read_numeric(-1, "int32_le")

def test_read_int64_negative_offset(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.read_numeric(-1, "int64_le")

def test_read_float32_negative_offset(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.read_numeric(-1, "float32_le")

def test_read_float64_negative_offset(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.read_numeric(-1, "float64_le")

def test_write_boundary_case1(filler_0_16):
    filler_0_16.write(0x0, b'hello')

def test_write_boundary_case2(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.write(0x10, b'hello')

def test_write_boundary_case3(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.write(-1, b'hello')

def test_read_boundary_case1(filler_0_16):
    assert filler_0_16.read(0x0, 5) == b"\x00\x00\x00\x00\x00"

def test_read_boundary_case2(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.read(0x10, 5)

def test_read_boundary_case3(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.read(-1, 5)

def test_read_uint8_array_boundary_case1(filler_0_16):
    assert filler_0_16.read_numeric_array(0, 10, "uint8_be") == (0, 0, 0, 0, 0, 0, 0, 0, 0, 0)

def test_read_uint8_array_boundary_case2(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.read_numeric_array(0x10, 10, "uint8_be") == (0, 0, 0, 0, 0, 0, 0, 0, 0, 0)

def test_read_uint8_array_boundary_case3(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.read_numeric_array(-1, 10, "uint8_be") == (0, 0, 0, 0, 0, 0, 0, 0, 0, 0)

def test_write_uint8_array_boundary_case1(filler_0_16):
    filler_0_16.write_numeric_array(0, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], "uint8_be")

def test_write_uint8_array_boundary_case2(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.write_numeric_array(0x10, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], "uint8_be")

def test_write_uint8_array_boundary_case3(filler_0_16):
    with pytest.raises(InvalidAddressError):
        filler_0_16.write_numeric_array(-1, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], "uint8_be")

@pytest.mark.skipif("NUMPY_SUPPORT == False")
def test_write_ndarray1():
    sec = Section(start_address = 0x1000, data = bytearray(32))
    arr = np.array([[11, 22, 33], [44, 55, 66]], dtype = "int32")
    sec.write_ndarray(0x1000, arr)
    assert sec.data == b'\x0b\x00\x00\x00\x16\x00\x00\x00!\x00\x00\x00,\x00\x00\x007\x00\x00\x00B\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'

@pytest.mark.skipif("NUMPY_SUPPORT == False")
def test_write_ndarray_out_of_bounds1():
    sec = Section(start_address = 0x1000, data = bytearray(32))
    arr = np.array([[11, 22, 33], [44, 55, 66]], dtype = "int64")
    with pytest.raises(InvalidAddressError):
        sec.write_ndarray(0x1000, arr)

@pytest.mark.skipif("NUMPY_SUPPORT == False")
def test_write_ndarray_out_of_bounds2():
    sec = Section(start_address = 0x1000, data = bytearray(32))
    arr = np.array([[11, 22, 33], [44, 55, 66]], dtype = "int64")
    with pytest.raises(InvalidAddressError):
        sec.write_ndarray(0x9ff, arr)

@pytest.mark.skipif("NUMPY_SUPPORT == False")
def test_read_ndarray_reshaped():
    sec = Section(start_address = 0x1000, data = bytearray(32))
    arr = np.array([[11, 22, 33], [44, 55, 66]], dtype = "int32")
    sec.write_ndarray(0x1000, arr)

    result =sec.read_ndarray(0x1000, 24, "int32_le", shape = (2, 3))
    assert np.array_equal(result, np.array([[11, 22, 33], [44, 55, 66]]))

@pytest.mark.skipif("NUMPY_SUPPORT == False")
def test_read_ndarray_flat():
    sec = Section(start_address = 0x1000, data = bytearray(32))
    arr = np.array([[11, 22, 33], [44, 55, 66]], dtype = "int32")
    sec.write_ndarray(0x1000, arr)

    result =sec.read_ndarray(0x1000, 24, "int32_le")
    assert np.array_equal(result, np.array([11, 22, 33, 44, 55, 66]))




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from array import array
import io
import unittest
import sys

import pytest

from objutils import dumps
from objutils.section import Section
from objutils.image import Image
from objutils.exceptions import InvalidAddressError
from objutils.utils import PYTHON_VERSION, create_string_buffer

class BaseTest(unittest.TestCase):

    def setUp(self):
        self.b0 = Image()
        self.b1 = Image()

    def tearDown(self):
        del self.b0
        del self.b1


class TestImageConstructors(unittest.TestCase):

    def testEmpty(self):
        img = Image()
        self.assertEqual(img.sections, [])

    def testNone(self):
        img = Image(None)
        self.assertEqual(img.sections, [])

    def testSingle(self):
        sec0 = Section(data = "hello", start_address = 0x100)
        img = Image(sec0)
        self.assertEqual(len(img.sections), 1)

    def testTwo(self):
        sec0 = Section(data = "hello", start_address = 0x100)
        sec1 = Section(data = "world", start_address = 0x200)
        img = Image((sec0, sec1))
        self.assertEqual(len(img.sections), 2)

    def testFailOnUnsupportedType(self):
        self.assertRaises(TypeError, Image, 4711)

@pytest.fixture
def images(scope = "module"):
    img0 = Image()
    img1 = Image()
    yield img0, img1
    del img0
    del img1

def test_insert1(images):
    img0, _ = images
    img0.insert_section(data = b"0123456789", start_address = 0x100)
    assert img0.sections[0].start_address == 0x100
    assert img0.sections[0].data == bytearray(b"0123456789")
    assert img0.sections[0].length == 10

def test_insert_overlapping1(images):
    img0, _ = images
    img0.insert_section(data = b"0123456789", start_address = 0x100)
    with pytest.raises(InvalidAddressError):
        img0.insert_section(data = b"0123456789", start_address = 0x100)

def test_insert_overlapping2(images):
    img0, _ = images
    img0.insert_section(data = b"0123456789", start_address = 0x100)
    with pytest.raises(InvalidAddressError):
        img0.insert_section(data = b"0123456789", start_address = 0x109)

def test_insert_overlapping3(images):
    img0, _ = images
    img0.insert_section(data = b"0123456789", start_address = 0x100)
    img0.insert_section(data = b"0123456789", start_address = 0x10a)

def test_insert_overlapping3(images):
    img0, _ = images
    img0.insert_section(data = b"0123456789", start_address = 0x100)
    with pytest.raises(InvalidAddressError):
        img0.insert_section(data = b"0123456789", start_address = 0x0f7)

def test_insert_overlapping4(images):
    img0, _ = images
    img0.insert_section(data = b"0123456789", start_address = 0x100)
    img0.insert_section(data = b"0123456789", start_address = 0x0f6)

def test_insert_overlapping5(images):
    img0, _ = images
    img0.insert_section(data = b"0123456789", start_address = 0x100)
    img0.insert_section(data = b"0123456789", start_address = 0x114)
    img0.insert_section(data = b"0123456789", start_address = 0x10a)

def test_update1(images):
    img0, _ = images
    img0.insert_section(data = b"0123456789", start_address = 0x100)
    with pytest.raises(InvalidAddressError):
        img0.update_section(data = b"abcdefghij", address = 0x10a)

def test_update2(images):
    img0, _ = images
    img0.insert_section(data = b"0123456789", start_address = 0x100)
    img0.update_section(data = b"abcdefghij", address = 0x109)

"""


    b0.hexdump(sys.stdout)
    captured = capsys.readouterr()
    assert captured.out == TEST1
"""
def test_sorting1(capsys):
    RES = """
Section #0000
-------------
00000010  30 31 32 33 34 35 36 37 38 39                    |0123456789      |
---------------
       10 bytes
---------------

Section #0001
-------------
00000020  30 31 32 33 34 35 36 37 38 39                    |0123456789      |
---------------
       10 bytes
---------------

Section #0002
-------------
00000030  30 31 32 33 34 35 36 37 38 39                    |0123456789      |
---------------
       10 bytes
---------------

Section #0003
-------------
00000040  30 31 32 33 34 35 36 37 38 39                    |0123456789      |
---------------
       10 bytes
---------------

Section #0004
-------------
00000050  30 31 32 33 34 35 36 37 38 39                    |0123456789      |
---------------
       10 bytes
---------------

Section #0005
-------------
00000060  30 31 32 33 34 35 36 37 38 39                    |0123456789      |
---------------
       10 bytes
---------------

Section #0006
-------------
00000070  30 31 32 33 34 35 36 37 38 39                    |0123456789      |
---------------
       10 bytes
---------------

Section #0007
-------------
00000080  30 31 32 33 34 35 36 37 38 39                    |0123456789      |
---------------
       10 bytes
---------------

Section #0008
-------------
00000090  30 31 32 33 34 35 36 37 38 39                    |0123456789      |
---------------
       10 bytes
---------------

Section #0009
-------------
00000100  30 31 32 33 34 35 36 37 38 39                    |0123456789      |
---------------
       10 bytes
---------------
"""
    img = Image(join = True)
    img.insert_section(data = b"0123456789", start_address = 0x100)
    img.insert_section(data = b"0123456789", start_address = 0x090)
    img.insert_section(data = b"0123456789", start_address = 0x080)
    img.insert_section(data = b"0123456789", start_address = 0x070)
    img.insert_section(data = b"0123456789", start_address = 0x060)
    img.insert_section(data = b"0123456789", start_address = 0x050)
    img.insert_section(data = b"0123456789", start_address = 0x040)
    img.insert_section(data = b"0123456789", start_address = 0x030)
    img.insert_section(data = b"0123456789", start_address = 0x020)
    img.insert_section(data = b"0123456789", start_address = 0x010)
    img.hexdump(sys.stdout)
    captured = capsys.readouterr()
    assert captured.out == RES

def test_sorting2(capsys):
    RES = """
Section #0000
-------------
00000010  30 31 32 33 34 35 36 37 38 39                    |0123456789      |
---------------
       10 bytes
---------------

Section #0001
-------------
00000020  30 31 32 33 34 35 36 37 38 39                    |0123456789      |
---------------
       10 bytes
---------------

Section #0002
-------------
00000030  30 31 32 33 34 35 36 37 38 39                    |0123456789      |
---------------
       10 bytes
---------------

Section #0003
-------------
00000040  30 31 32 33 34 35 36 37 38 39                    |0123456789      |
---------------
       10 bytes
---------------

Section #0004
-------------
00000050  30 31 32 33 34 35 36 37 38 39                    |0123456789      |
---------------
       10 bytes
---------------

Section #0005
-------------
00000060  30 31 32 33 34 35 36 37 38 39                    |0123456789      |
---------------
       10 bytes
---------------

Section #0006
-------------
00000070  30 31 32 33 34 35 36 37 38 39                    |0123456789      |
---------------
       10 bytes
---------------

Section #0007
-------------
00000080  30 31 32 33 34 35 36 37 38 39                    |0123456789      |
---------------
       10 bytes
---------------

Section #0008
-------------
00000090  30 31 32 33 34 35 36 37 38 39                    |0123456789      |
---------------
       10 bytes
---------------

Section #0009
-------------
00000100  30 31 32 33 34 35 36 37 38 39                    |0123456789      |
---------------
       10 bytes
---------------
"""
    img = Image(join = False)
    img.insert_section(data = b"0123456789", start_address = 0x100)
    img.insert_section(data = b"0123456789", start_address = 0x090)
    img.insert_section(data = b"0123456789", start_address = 0x080)
    img.insert_section(data = b"0123456789", start_address = 0x070)
    img.insert_section(data = b"0123456789", start_address = 0x060)
    img.insert_section(data = b"0123456789", start_address = 0x050)
    img.insert_section(data = b"0123456789", start_address = 0x040)
    img.insert_section(data = b"0123456789", start_address = 0x030)
    img.insert_section(data = b"0123456789", start_address = 0x020)
    img.insert_section(data = b"0123456789", start_address = 0x010)
    img.hexdump(sys.stdout)
    captured = capsys.readouterr()
    assert captured.out == RES

@pytest.mark.skip
def test_sorting7(capsys):
    RES = """
Section #0000
-------------
00000000  00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f  |................|
---------------
       16 bytes
---------------

Section #0001
-------------
00000010  00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f  |................|
---------------
       16 bytes
---------------

Section #0002
-------------
00000020  00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f  |................|
---------------
       16 bytes
---------------

Section #0003
-------------
00000030  00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f  |................|
---------------
       16 bytes
---------------
"""
    img = Image(join = False)
    img.insert_section(data = range(16), start_address = 0x030)
    img.insert_section(data = range(16), start_address = 0x020)
    img.insert_section(data = range(16), start_address = 0x010)
    img.insert_section(data = range(16), start_address = 0x000)
    img.hexdump(sys.stdout)
    captured = capsys.readouterr()
    assert captured.out == RES

def test_sorting8(capsys):
    RES = """
Section #0000
-------------
00000000  00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f  |................|
          *
00000030  00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f  |................|
---------------
       64 bytes
---------------
"""
    img = Image(join = True)
    img.insert_section(data = range(16), start_address = 0x030)
    img.insert_section(data = range(16), start_address = 0x020)
    img.insert_section(data = range(16), start_address = 0x010)
    img.insert_section(data = range(16), start_address = 0x000)
    img.hexdump(sys.stdout)
    captured = capsys.readouterr()
    assert captured.out == RES

@pytest.mark.skip
def test_get_section1():
    img = Image(join = False)
    img.insert_section(data = range(16), start_address = 0x030)
    img.insert_section(data = range(16), start_address = 0x020)
    img.insert_section(data = range(16), start_address = 0x010)
    img.insert_section(data = range(16), start_address = 0x000)
    sec = img.get_section(0x33)
    assert sec.start_address == 0x30
    sec = img.get_section(0x22)
    assert sec.start_address == 0x20
    sec = img.get_section(0x11)
    assert sec.start_address == 0x10
    sec = img.get_section(0x02)
    assert sec.start_address == 0x00

def test_get_section2():
    img = Image(join = True)
    img.insert_section(data = range(16), start_address = 0x030)
    img.insert_section(data = range(16), start_address = 0x020)
    img.insert_section(data = range(16), start_address = 0x010)
    img.insert_section(data = range(16), start_address = 0x000)
    sec = img.get_section(0x33)
    assert sec.start_address == 0x00
    assert len(sec) == 64


@pytest.mark.parametrize("join,", [False, True])
def test_get_section_raises(join):
    img = Image(join = join)
    img.insert_section(data = range(10), start_address = 0x030)
    img.insert_section(data = range(10), start_address = 0x020)
    img.insert_section(data = range(10), start_address = 0x010)
    img.insert_section(data = range(10), start_address = 0x000)
    with pytest.raises(InvalidAddressError):
        sec = img.get_section(0x3a)
    with pytest.raises(InvalidAddressError):
        sec = img.get_section(0x2b)
    with pytest.raises(InvalidAddressError):
        sec = img.get_section(0x1c)
    with pytest.raises(InvalidAddressError):
        sec = img.get_section(0x0d)


class Equality(BaseTest):

    def testEqualImagesShallCompareEqualCase1(self):
        self.b0.insert_section("01234567890", 0x1000)
        self.b1.insert_section("01234567890", 0x1000)
        self.assertTrue(self.b0 == self.b1)

    def testEqualImagesShallCompareEqualCase2(self):
        self.b0.insert_section("01234567890", 0x1000)
        self.b1.insert_section("01234567890", 0x1000)
        self.assertFalse(self.b0 != self.b1)


class TestCreateSections(BaseTest):

    SREC = b"S1131000000102030405060708090A0B0C0D0E0F64"

    def runSectionTestPass(self, data):
        self.b0.insert_section(data, 0x1000)
        result = dumps('srec', self.b0)
        self.assertEqual(result, self.SREC)

    def runSectionTestFail(self, data):
        self.assertRaises(ValueError, self.b0.insert_section, data, 0x1000)

    def testCreateSectionFromStringWorks(self):
        self.runSectionTestPass('\x00\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f')

    def testCreateSectionFromListOfBytesWorks(self):
        self.runSectionTestPass([0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f])

    def testCreateSectionFromListOfWordsFails(self):
        self.runSectionTestFail([0x0001, 0x0203, 0x0405, 0x0607, 0x0809, 0x0a0b, 0x0c0d, 0x0e0f])

    def testCreateSectionFromRangeWorks(self):
        self.runSectionTestPass(range(16))

    def testEnsureSectionHasCopySemantics(self):
        data = [0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f]
        self.b0.insert_section(data, 0x1000)
        data.extend([0x10, 0x20, 0x30, 0x40])
        result = dumps('srec', self.b0)
        self.assertEqual(result, self.SREC)

    def testEmptySectionProducesEmptiness(self):
        self.b0.insert_section([], 0x1000)
        result = dumps('srec', self.b0)
        self.assertEqual(result, b'')

    def testCreateSectionFromBytesWorks(self):
        if PYTHON_VERSION.major == 3:
            self.runSectionTestPass(bytes([0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f]))
        else:
            pass    # Unfortunately, this test-case isn't supported by Python 2.x

    def testCreateSectionFromByteArrayWorks(self):
        self.runSectionTestPass(bytearray([0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f]))

    def testCreateSectionFromArrayBWorks(self):
        self.runSectionTestPass(array('B', [0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f]))

    def testOmittedAddressYieldsToZero(self):
        self.b0.insert_section(range(16))
        data = self.b0
        self.assertEqual(data[0].start_address, 0X00000000)

    #def testCreateSectionFromArrayHWorks(self):
    #    if sys.byteorder == 'little':
    #        self.runSectionTestPass(array('H', [0x0100, 0x0302, 0x0504, 0x0706, 0x0908, 0x0b0a, 0x0d0c, 0x0f0e]))
    #    else:
    #        self.runSectionTestPass(array('H', [0x0001, 0x0203, 0x0405, 0x0607, 0x0809, 0x0a0b, 0x0c0d, 0x0e0f]))

    #def testCreateSectionFromUnicodeFails(self):
    #    self.runSectionTestPass('\x00\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f')

class TestImageSlices(BaseTest):

    def createImage(self):
        self.b0.insert_section(range(16), 0x1000)
        self.b0.insert_section(range(16), 0x2000)
        self.b0.insert_section(range(16), 0x3000)

    def testLenWorks(self):
        self.createImage()
        image = self.b0
        self.assertEqual(len(image), 48)

    def testSlicingWorks(self):
        self.createImage()
        image = self.b0
        section = image[1]
        self.assertTrue(isinstance(section, Section))

    def testIteration(self):
        self.createImage()
        image = self.b0
        for section in image:
            self.assertTrue(isinstance(section, Section))

    def testHexdumpWorksOnSlice(self):
        self.createImage()
        image = self.b0
        section = image[1]
        if PYTHON_VERSION.major == 3:
            buf = io.TextIOWrapper(create_string_buffer())
        else:
            buf = create_string_buffer()
        section.hexdump(buf)

def test_read_write_datatypes_require_suffix():
    img = Image(Section(data = bytearray(32), start_address = 0x1000))
    with pytest.raises(TypeError):
        img.read_numeric(0x1000, "uint8")
    with pytest.raises(TypeError):
        img.read_numeric(0x1000, "int8")
    with pytest.raises(TypeError):
        img.read_numeric(0x1000, "uint16")
    with pytest.raises(TypeError):
        img.read_numeric(0x1000, "int16")
    with pytest.raises(TypeError):
        img.read_numeric(0x1000, "uint32")
    with pytest.raises(TypeError):
        img.read_numeric(0x1000, "int32")
    with pytest.raises(TypeError):
        img.read_numeric(0x1000, "uint64")
    with pytest.raises(TypeError):
        img.read_numeric(0x1000, "int64")
    with pytest.raises(TypeError):
        img.read_numeric(0x1000, "float32")
    with pytest.raises(TypeError):
        img.read_numeric(0x1000, "float64")

def test_invalid_read_write_datatype_raises():
    img = Image(Section(data = bytearray(32), start_address = 0x1000))
    with pytest.raises(TypeError):
        img.read_numeric(0x1000, "foobar_le")

def test_read_write():
    img = Image(Section(data = bytearray(32), start_address = 0x1000))
    img.write(0x1000, b'hello')
    assert img.read(0x1000, 5) == b"hello"

def test_write_boundary_case1():
    img = Image(Section(data = bytearray(5), start_address = 0x1000))
    img.write(0x1000, b'hello')

def test_write_boundary_case2():
    img = Image(Section(data = bytearray(5), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.write(0x1001, b'hello')

def test_write_boundary_case3():
    img = Image(Section(data = bytearray(5), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.write(0x0fff, b'hello')

def test_read_boundary_case1():
    img = Image(Section(data = bytearray(5), start_address = 0x1000))
    assert img.read(0x1000, 5) == b"\x00\x00\x00\x00\x00"

def test_read_boundary_case2():
    img = Image(Section(data = bytearray(5), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.read(0x1001, 5)

def test_read_boundary_case3():
    img = Image(Section(data = bytearray(5), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.read(0x0fff, 5)

def test_read_write_uint8():
    img = Image(Section(data = bytearray(32), start_address = 0x1000))
    img.write_numeric(0x1000, 0x55, "uint8_be")
    assert img.read_numeric(0x1000, "uint8_be") == 0x55
    img.write_numeric(0x1000, 0x55, "uint8_le")
    assert img.read_numeric(0x1000, "uint8_le") == 0x55

def test_read_write_int8():
    img = Image(Section(data = bytearray(32), start_address = 0x1000))
    img.write_numeric(0x1000, -20, "int8_be")
    assert img.read_numeric(0x1000, "int8_be") == -20
    img.write_numeric(0x1000, -20, "int8_le")
    assert img.read_numeric(0x1000, "int8_le") == -20

def test_read_write_uint16():
    img = Image(Section(data = bytearray(32), start_address = 0x1000))
    img.write_numeric(0x1000, 0x5555, "uint16_be")
    assert img.read_numeric(0x1000, "uint16_be") == 0x5555
    img.write_numeric(0x1000, 0x5555, "uint16_le")
    assert img.read_numeric(0x1000, "uint16_le") == 0x5555

def test_read_write_int16():
    img = Image(Section(data = bytearray(32), start_address = 0x1000))
    img.write_numeric(0x1000, -2000, "int16_be")
    assert img.read_numeric(0x1000, "int16_be") == -2000
    img.write_numeric(0x1000, -2000, "int16_le")
    assert img.read_numeric(0x1000, "int16_le") == -2000

def test_read_write_uint32():
    img = Image(Section(data = bytearray(32), start_address = 0x1000))
    img.write_numeric(0x1000, 0x55555555, "uint32_be")
    assert img.read_numeric(0x1000, "uint32_be") == 0x55555555
    img.write_numeric(0x1000, 0x55555555, "uint32_le")
    assert img.read_numeric(0x1000, "uint32_le") == 0x55555555

def test_read_write_int32():
    img = Image(Section(data = bytearray(32), start_address = 0x1000))
    img.write_numeric(0x1000, -20000000, "int32_be")
    assert img.read_numeric(0x1000, "int32_be") == -20000000
    img.write_numeric(0x1000, -20000000, "int32_le")
    assert img.read_numeric(0x1000, "int32_le") == -20000000

def test_read_write_uint64():
    img = Image(Section(data = bytearray(32), start_address = 0x1000))
    img.write_numeric(0x1000, 0x5555555555555555, "uint64_be")
    assert img.read_numeric(0x1000, "uint64_be") == 0x5555555555555555
    img.write_numeric(0x1000, 0x5555555555555555, "uint64_le")
    assert img.read_numeric(0x1000, "uint64_le") == 0x5555555555555555

def test_read_write_int64():
    img = Image(Section(data = bytearray(32), start_address = 0x1000))
    img.write_numeric(0x1000, -200000000000, "int64_be")
    assert img.read_numeric(0x1000, "int64_be") == -200000000000
    img.write_numeric(0x1000, -200000000000, "int64_le")
    assert img.read_numeric(0x1000, "int64_le") == -200000000000

def test_read_write_float32():
    img = Image(Section(data = bytearray(32), start_address = 0x1000))
    img.write_numeric(0x1000, 1234.5678, "float32_be")
    assert img.read_numeric(0x1000, "float32_be") == 1234.5677490234375
    img.write_numeric(0x1000, 1234.5678, "float32_le")
    assert img.read_numeric(0x1000, "float32_le") == 1234.5677490234375

def test_read_write_float64():
    img = Image(Section(data = bytearray(32), start_address = 0x1000))
    img.write_numeric(0x1000, 1234.5678, "float64_be")
    assert img.read_numeric(0x1000, "float64_be") == 1234.5678
    img.write_numeric(0x1000, 1234.5678, "float64_le")
    assert img.read_numeric(0x1000, "float64_le") == 1234.5678

def test_read_write_string():
    img = Image(Section(data = bytearray(32), start_address = 0x1000))
    img.write_string(0x1000, 'hello')
    assert img.read_string(0x1000) == "hello"

def test_read_write_string_length():
    img = Image(Section(data = bytearray(32), start_address = 0x1000))
    img.write_string(0x1000, 'hello world!!!')
    assert img.read_string(0x1000, length = 5) == "hello"

def test_read_unterminated_string_raises():
    img = Image(Section(data = bytearray(b"\x0a" * 32), start_address = 0x1000))
    img.write(0x1000, b'hello')
    with pytest.raises(TypeError):
        data = img.read_string(0x1000)
##
def test_write_array_data_must_be_iterable():
    img = Image(Section(data = bytearray(32), start_address = 0x1000))
    with pytest.raises(TypeError):
        img.write_numeric_array(0x1000, 0x55, "uint8_be")

def test_write_uint8_array_boundary_case1():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    img.write_numeric_array(0x1000, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], "uint8_be")
    assert img[0].length == 10

def test_write_uint8_array_boundary_case2():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    img.write_numeric_array(0x1005, [1, 2, 3, 4, 5], "uint8_be")
    assert img[0].length == 10

def test_write_uint8_array_boundary_case3():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.write_numeric_array(0x1006, [1, 2, 3, 4, 5], "uint8_be")

def test_write_uint8_array_boundary_case4():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.write_numeric_array(0x0fff, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], "uint8_be")

def test_read_uint8_array_boundary_case1():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    assert img.read_numeric_array(0x1000, 10, "uint8_be") == (0, 0, 0, 0, 0, 0, 0, 0, 0, 0)

def test_read_uint8_array_boundary_case2():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    assert img.read_numeric_array(0x1005, 5, "uint8_be") == (0, 0, 0, 0, 0)

def test_read_uint8_array_boundary_case3():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.read_numeric_array(0x1006, 5, "uint8_be") == (0, 0, 0, 0, 0)

def test_read_uint8_array_boundary_case4():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.read_numeric_array(0x0fff, 5, "uint8_be") == (0, 0, 0, 0, 0)

def test_read_write_uint8_array():
    img = Image(Section(data = bytearray(32), start_address = 0x1000))
    img.write_numeric_array(0x1000, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], "uint8_be")
    assert img[0].length == 32
    assert img.read_numeric_array(0x1000, 10, "uint8_be") == (1, 2, 3, 4, 5, 6, 7, 8, 9, 10)
    img.write_numeric_array(0x1000, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], "uint8_le")
    assert img[0].length == 32
    assert img.read_numeric_array(0x1000, 10, "uint8_le") == (1, 2, 3, 4, 5, 6, 7, 8, 9, 10)

def test_read_write_int8_array():
    img = Image(Section(data = bytearray(32), start_address = 0x1000))
    img.write_numeric_array(0x1000, [-1, -2, -3, -4, -5, -6, -7, -8, -9, -10], "int8_be")
    assert img[0].length == 32
    assert img.read_numeric_array(0x1000, 10, "int8_be") == (-1, -2, -3, -4, -5, -6, -7, -8, -9, -10)
    img.write_numeric_array(0x1000, [-1, -2, -3, -4, -5, -6, -7, -8, -9, -10], "int8_le")
    assert img[0].length == 32
    assert img.read_numeric_array(0x1000, 10, "int8_le") == (-1, -2, -3, -4, -5, -6, -7, -8, -9, -10)

def test_read_write_uint16_array():
    img = Image(Section(data = bytearray(32), start_address = 0x1000))
    img.write_numeric_array(0x1000, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], "uint16_be")
    assert img[0].length == 32
    assert img.read_numeric_array(0x1000, 10, "uint16_be") == (1, 2, 3, 4, 5, 6, 7, 8, 9, 10)
    img.write_numeric_array(0x1000, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], "uint16_le")
    assert img[0].length == 32
    assert img.read_numeric_array(0x1000, 10, "uint16_le") == (1, 2, 3, 4, 5, 6, 7, 8, 9, 10)

def test_read_write_int16_array():
    img = Image(Section(data = bytearray(32), start_address = 0x1000))
    img.write_numeric_array(0x1000, [-1, -2, -3, -4, -5, -6, -7, -8, -9, -10], "int16_be")
    assert img[0].length == 32
    assert img.read_numeric_array(0x1000, 10, "int16_be") == (-1, -2, -3, -4, -5, -6, -7, -8, -9, -10)
    img.write_numeric_array(0x1000, [-1, -2, -3, -4, -5, -6, -7, -8, -9, -10], "int16_le")
    assert img[0].length == 32
    assert img.read_numeric_array(0x1000, 10, "int16_le") == (-1, -2, -3, -4, -5, -6, -7, -8, -9, -10)

def test_read_write_uint32_array():
    img = Image(Section(data = bytearray(64), start_address = 0x1000))
    img.write_numeric_array(0x1000, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], "uint32_be")
    assert img[0].length == 64
    assert img.read_numeric_array(0x1000, 10, "uint32_be") == (1, 2, 3, 4, 5, 6, 7, 8, 9, 10)
    img.write_numeric_array(0x1000, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], "uint32_le")
    assert img[0].length == 64
    assert img.read_numeric_array(0x1000, 10, "uint32_le") == (1, 2, 3, 4, 5, 6, 7, 8, 9, 10)

def test_read_write_int32_array():
    img = Image(Section(data = bytearray(64), start_address = 0x1000))
    img.write_numeric_array(0x1000, [-1, -2, -3, -4, -5, -6, -7, -8, -9, -10], "int32_be")
    assert img[0].length == 64
    assert img.read_numeric_array(0x1000, 10, "int32_be") == (-1, -2, -3, -4, -5, -6, -7, -8, -9, -10)
    img.write_numeric_array(0x1000, [-1, -2, -3, -4, -5, -6, -7, -8, -9, -10], "int32_le")
    assert img[0].length == 64
    assert img.read_numeric_array(0x1000, 10, "int32_le") == (-1, -2, -3, -4, -5, -6, -7, -8, -9, -10)

def test_read_write_uint64_array():
    img = Image(Section(data = bytearray(128), start_address = 0x1000))
    img.write_numeric_array(0x1000, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], "uint64_be")
    assert img[0].length == 128
    assert img.read_numeric_array(0x1000, 10, "uint64_be") == (1, 2, 3, 4, 5, 6, 7, 8, 9, 10)
    img.write_numeric_array(0x1000, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], "uint64_le")
    assert img[0].length == 128
    assert img.read_numeric_array(0x1000, 10, "uint64_le") == (1, 2, 3, 4, 5, 6, 7, 8, 9, 10)

def test_read_write_int64_array():
    img = Image(Section(data = bytearray(128), start_address = 0x1000))
    img.write_numeric_array(0x1000, [-1, -2, -3, -4, -5, -6, -7, -8, -9, -10], "int64_be")
    assert img[0].length == 128
    assert img.read_numeric_array(0x1000, 10, "int64_be") == (-1, -2, -3, -4, -5, -6, -7, -8, -9, -10)
    img.write_numeric_array(0x1000, [-1, -2, -3, -4, -5, -6, -7, -8, -9, -10], "int64_le")
    assert img[0].length == 128
    assert img.read_numeric_array(0x1000, 10, "int64_le") == (-1, -2, -3, -4, -5, -6, -7, -8, -9, -10)

def test_read_write_float32_array():
    img = Image(Section(data = bytearray(64), start_address = 0x1000))
    img.write_numeric_array(0x1000, [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0], "float32_be")
    assert img[0].length == 64
    assert img.read_numeric_array(0x1000, 10, "float32_be") == (1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0)
    img.write_numeric_array(0x1000, [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0], "float32_le")
    assert img[0].length == 64
    assert img.read_numeric_array(0x1000, 10, "float32_le") == (1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0)

def test_read_write_float64_array():
    img = Image(Section(data = bytearray(128), start_address = 0x1000))
    img.write_numeric_array(0x1000, [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0], "float64_be")
    assert img[0].length == 128
    assert img.read_numeric_array(0x1000, 10, "float64_be") == (1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0)
    img.write_numeric_array(0x1000, [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0], "float64_le")
    assert img[0].length == 128
    assert img.read_numeric_array(0x1000, 10, "float64_le") == (1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0)

#
def test_write_uint8_negative_offset():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.write_numeric(0x0fff, 0xff, "uint8_le")

def test_write_uint16_negative_offset():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.write_numeric(0x0fff, 0xff, "uint16_le")

def test_write_uint32_negative_offset():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.write_numeric(0x0fff, 0xff, "uint32_le")

def test_write_uint64_negative_offset():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.write_numeric(0x0fff, 0xff, "uint64_le")

def test_write_int8_negative_offset():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.write_numeric(0x0fff, 0xff, "int8_le")

def test_write_int16_negative_offset():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.write_numeric(0x0fff, 0xff, "int16_le")

def test_write_int32_negative_offset():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.write_numeric(0x0fff, 0xff, "int32_le")

def test_write_int64_negative_offset():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.write_numeric(0x0fff, 0xff, "int64_le")

def test_write_float32_negative_offset():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.write_numeric(0x0fff, 3.14159, "float32_le")

def test_write_float64_negative_offset():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.write_numeric(0x0fff, 3.14159, "float64_le")

def test_read_uint8_negative_offset():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.read_numeric(0x0fff, "uint8_le")

def test_read_uint16_negative_offset():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.read_numeric(0x0fff, "uint16_le")

def test_read_uint32_negative_offset():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.read_numeric(0x0fff, "uint32_le")

def test_read_uint64_negative_offset():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.read_numeric(0x0fff, "uint64_le")

def test_read_int8_negative_offset():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.read_numeric(0x0fff, "int8_le")

def test_read_int16_negative_offset():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.read_numeric(0x0fff, "int16_le")

def test_read_int32_negative_offset():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.read_numeric(0x0fff, "int32_le")

def test_read_int64_negative_offset():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.read_numeric(0x0fff, "int64_le")

def test_read_float32_negative_offset():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.read_numeric(0x0fff, "float32_le")

def test_read_float64_negative_offset():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.read_numeric(0x0fff, "float64_le")

def test_read_uint8_array_boundary_case1():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    assert img.read_numeric_array(0x1000, 10, "uint8_be") == (0, 0, 0, 0, 0, 0, 0, 0, 0, 0)

def test_read_uint8_array_boundary_case2():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.read_numeric_array(0x1001, 10, "uint8_be") == (0, 0, 0, 0, 0, 0, 0, 0, 0, 0)

def test_read_uint8_array_boundary_case3():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.read_numeric_array(0x0fff, 10, "uint8_be") == (0, 0, 0, 0, 0, 0, 0, 0, 0, 0)

def test_write_uint8_array_boundary_case1():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    img.write_numeric_array(0x1000, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], "uint8_be")

def test_write_uint8_array_boundary_case2():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.write_numeric_array(0x1001, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], "uint8_be")

def test_write_uint8_array_boundary_case3():
    img = Image(Section(data = bytearray(10), start_address = 0x1000))
    with pytest.raises(InvalidAddressError):
        img.write_numeric_array(0x0fff, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], "uint8_be")

if __name__ == '__main__':
    unittest.main()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.1.1"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2019 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from collections import defaultdict, namedtuple
from functools import partial
import math
import os
import re

from objutils.section import Section, join_sections
from objutils.image import Image
from operator import itemgetter
from objutils.utils import slicer, create_string_buffer, PYTHON_VERSION
from objutils.logger import Logger


'''
MemoryBlocks
------------
Segment-Types: generic,code,const,bss,... (s. ELF!?)
'''

"""
TODO:   Create a base class for Reader/Writer common functionality.
        FactoryMethods ???
        Parameter ['exceptionsOnErrors'].
"""

SIXTEEN_BITS    = 0
TWENTY_BITS     = 1
TWENTYFOUR_BITS = 2
THIRTYTWO_BITS  = 3

START           = 0
LENGTH          = 1
TYPE            = 2
ADDRESS         = 3
DATA            = 4
UNPARSED        = 5
CHECKSUM        = 6
ADDR_CHECKSUM   = 7

MAP_GROUP_TO_REGEX = {
    LENGTH          : "(?P<length>[0-9a-zA-Z]{%d})",
    TYPE            : "(?P<type>\d{%d})",
    ADDRESS         : "(?P<address>[0-9a-zA-Z]{%d})",
    DATA            : "(?P<chunk>[0-9a-zA-Z]+)",
    UNPARSED        : "(?P<chunk>.*)",
    CHECKSUM        : "(?P<checksum>[0-9a-zA-Z]{%d})",
    ADDR_CHECKSUM   : "(?P<addrChecksum>[0-9a-zA-Z]{%d})",
}

MAP_CHAR_TO_GROUP = {
    'L': LENGTH,
    'T': TYPE,
    'A': ADDRESS,
    'D': DATA,
    'U': UNPARSED,
    'C': CHECKSUM,
    'B': ADDR_CHECKSUM,
}

TYPE_FROM_RECORD = 0

atoi = partial(int, base = 16)

class Invalidrecord_typeError(Exception): pass
class InvalidRecordLengthError(Exception): pass
class InvalidRecordChecksumError(Exception): pass
class AddressRangeToLargeError(Exception): pass

MetaRecord = namedtuple('MetaRecord', 'format_type address chunk')

class FormatParser(object):
    def __init__(self, fmt, data_separator = None):
        self.fmt = fmt
        self.translated_format = []
        self.data_separator = data_separator

    def parse(self):
        group = ''
        prevCh = ''
        for ch in self.fmt:
            if ch != prevCh:
                if group != '':
                    self.translateFormat(group)
                    group = ''
            group += ch
            prevCh = ch
        self.translateFormat(group)
        self.translated_format.append((0, 0, r"(?P<junk>(.*?))$"))  #??
        ft = "^{0}".format(''.join(map(itemgetter(2), self.translated_format)))
        return re.compile(ft, re.DOTALL | re.MULTILINE)

    def translateFormat(self, group):
        group_number = MAP_CHAR_TO_GROUP.get(group[0])
        length = len(group)
        if group_number is None:    # Handle invariants (i.e. fixed chars).
            if group[0] == ' ':
                expr = '\s{{{0!s}}}'.format(length )
            else:
                expr = group[0] * length
        else:
            expr = MAP_GROUP_TO_REGEX.get(group_number)
            if group_number == START:
                expr = expr % (self.startSign, )
            elif group_number == DATA:
                "(?P<chunk>[0-9a-zA-Z]*)"
                if self.data_separator is not None:
                    expr = "(?P<chunk>[0-9a-zA-Z{0!s}]*)".format(self.data_separator )
                else:
                    pass
            elif group_number == UNPARSED:
                #print expr
                pass
            else:
                expr = expr % (length, )
        self.translated_format.append((group_number, length, expr))


class Container(object):
    def __init__(self):
        self.processing_instructions = []

    def add_processing_instruction(self, pi):
        self.processing_instructions.append(pi)


class BaseType(object):

    def error(self, msg):
        self.logger.error(msg)
        self.valid = False

    def warn(self, msg):
        self.logger.warn(msg)

    def info(self, msg):
        self.logger.info(msg)

    def debug(self, msg):
        self.logger.debug(msg)


class Reader(BaseType):
    ALIGMENT = 0  # 2**n
    DATA_SEPARATOR = None
    VALID_CHARS = re.compile(r"^[a-fA-F0-9 :/;,%\n\r!?S]*$") # General case, fits most formats.

    def __init__(self):
        self.logger = Logger("Reader")
        if isinstance(self.FORMAT_SPEC, str):
            self.formats = [FormatParser(self.FORMAT_SPEC, self.DATA_SEPARATOR).parse()]
        elif isinstance(self.FORMAT_SPEC, (list, tuple)):
            self.formats = []
            for format_type, format in self.FORMAT_SPEC:
                self.formats.append((format_type, FormatParser(format, self.DATA_SEPARATOR).parse()))

    def load(self, fp, **kws):
        if isinstance(fp, str):
            fp = open(fp, "rb")
        if PYTHON_VERSION.major == 3:
            data = self.read(fp)#.decode()
            if hasattr(fp, "close"):
                fp.close()
            return data
        else:
            data = self.read(fp)
            if hasattr(fp, "close"):
                fp.close()
            return data


    def loads(self, image, **kws):
        if PYTHON_VERSION.major == 3:
            if isinstance(image, str):
                return self.load(create_string_buffer(bytes(image, "ascii")))
            else:
                return self.load(create_string_buffer(image))
        else:
            return self.load(create_string_buffer(image))

    def read(self, fp):
        sections = []
        matched = False
        self.valid = True
        meta_data = defaultdict(list)
        for (line_number, line) in enumerate(fp.readlines(), 1):
            for format_type, format in self.formats:
                if isinstance(line, bytes):
                    match = format.match(line.decode())
                else:
                    match = format.match(line)
                if match:
                    matched = True
                    container = Container()
                    container.line_number = line_number
                    dict_ = match.groupdict()
                    if dict_ != {}:
                        # Handle scalar values.
                        for key, value in dict_.items():
                            if key not in ('chunk', 'junk'):
                                setattr(container, key, atoi(value))
                            elif key == 'junk':
                                setattr(container, key, value)
                        if 'chunk' in dict_:
                            if self.parseData(container, format_type):
                                chunk = bytearray.fromhex(dict_['chunk'])
                            else:
                                # don't convert/parse stuff like symbols.
                                chunk = dict_['chunk']
                            dict_.pop('chunk')
                            setattr(container, 'chunk', chunk)
                        self.check_line(container, format_type)
                        # this is to handle esoteric stuff like Intel seg:offs addressing and symbols.
                        self.special_processing(container, format_type)
                        if self.is_data_line(container, format_type):
                            sections.append(Section(container.address, container.chunk))
                        else:
                            chunk = container.chunk if hasattr(container, 'chunk') else None
                            address = container.address if hasattr(container, 'address') else None
                            meta_data[format_type].append(MetaRecord(format_type, address, chunk))
                    break
            if not matched:
                self.warn("Ignoring garbage line #{0:d}".format(line_number))
        if sections:
            return Image(sections = join_sections(sections), meta = meta_data)
        else:
            self.error("File seems to be invalid.")
            return Image([], valid = False)

    def _address_space(self, value):
        if value < 2**16:
            return SIXTEEN_BITS
        elif value < 2**20:
            return TWENTY_BITS
        elif value < 2**24:
            return TWENTYFOUR_BITS
        elif value < 2**32:
            return THIRTYTWO_BITS
        else:
            raise ValueError("Unsupported Addressspace size.")

    def maybe_binary_file(self, fp):
        fp.seek(0, os.SEEK_SET)
        header = fp.read(128)
        fp.seek(0, os.SEEK_SET)
        result = not bool(self.VALID_CHARS.match(header.decode()))
        return result

    def probe(self, fp):
        "Determine if object is valid."
        if self.maybe_binary_file(fp):
            return False
        matched = False
        for (line_number, line) in enumerate(fp.readlines(), 1):
            for format_type, format in self.formats: # NOTE: Same as in 'read()'!
                if isinstance(line, bytes):
                    match = format.match(line.decode())
                else:
                    match = format.match(line)
                if match:
                    matched = True
                    break
            if matched or line_number > 3:
                break
        return matched

    def probes(self, image):
        if PYTHON_VERSION.major == 3:
            if isinstance(image, str):
                return self.probe(create_string_buffer(bytes(image, "ascii")))
            else:
                return self.probe(create_string_buffer(image))
        else:
            return self.probe(create_string_buffer(image))

    def check_line(self, line, format_type):
        raise NotImplementedError()

    def is_data_line(self, line, format_type):
        raise NotImplementedError()

    def classifyLine(self, line):
        raise NotImplementedError()

    def special_processing(self, line, format_type):
        pass

    def parseData(self, line, format_type):
        return True


class Writer(BaseType):

    def __init__(self):
        self.logger = Logger("Writer")

    def dump(self, fp, image, row_length = 16, **kws):   # TODO: rename to bytesPerRow!
        if isinstance(fp, str):
            fp = open(fp, "wb")
        fp.write(self.dumps(image, row_length))
        if hasattr(fp, "close"):
            fp.close()

    def dumps(self, image, row_length = 16, **kws):
        result = []
        self.row_length = row_length
        if hasattr(image, "sections") and not image.sections:
            return b''
        if self.calculate_address_bits(image) > self.MAX_ADDRESS_BITS:
            raise AddressRangeToLargeError('could not encode image.')
        params = self.set_parameters(**kws)
        self.pre_processing(image)
        header = self.compose_header(image.meta if hasattr(image, 'meta') else {})
        if header:
            result.append(header)
        for section in image:
            address = section.start_address
            rows = slicer(section.data, row_length, lambda x:  [int(y) for y in x])
            for row in rows:
                length = len(row)
                result.append(self.compose_row(address, length, row))
                address += row_length
        footer = self.compose_footer(image.meta if hasattr(image, 'meta') else {})
        if footer:
            result.append(footer)
        if PYTHON_VERSION.major == 3:
            return self.post_processing(bytes('\n'.join(result), "ascii"))
        else:
            return self.post_processing(bytes('\n'.join(result)))

    def calculate_address_bits(self, image):
        if hasattr(image, "sections"):
            last_segment = sorted(image.sections, key = lambda s: s.start_address)[-1]
        else:
            last_segment = image
        highest_address = last_segment.start_address + last_segment.length
        return int(math.ceil(math.log(highest_address + 1) / math.log(2)))

    def post_processing(self, data):
        return data

    def pre_processing(self, image):
        pass

    def set_parameters(self, **kws):
        params = {}
        for k, v in kws.items():
            try:
                params[k] = getattr(self, k)
            except AttributeError:
                raise AttributeError("Invalid keyword argument '{0!s}'.".format(k))
            else:
                setattr(self, k, v)
        return params

    def compose_row(self, address, length, row):
        raise NotImplementedError()

    def compose_header(self, meta):
        return None

    def compose_footer(self, meta):
        return None

    def word_to_bytes(self, word):
        word = int(word)
        h = (word & 0xff00) >> 8
        l = word & 0x00ff
        return h, l

    @staticmethod
    def hex_bytes(row, spaced = False):
        spacer = ' ' if spaced else ''
        return spacer.join(["{0:02X}".format(x) for x in row])


class ASCIIHexReader(Reader):

    FORMAT_SPEC = None

    def __init__(self, address_pattern, data_pattern, etx_pattern, separators = ', '):
        self.separators = separators
        self.DATA_PATTERN = re.compile(data_pattern.format(separators), re.DOTALL | re.MULTILINE)
        self.ADDRESS_PATTERN = re.compile(address_pattern, re.DOTALL | re.MULTILINE)
        self.ETX_PATTERN = re.compile(etx_pattern, re.DOTALL | re.MULTILINE)
        self.SPLITTER = re.compile('[{0}]'.format(separators))
        self.patterns = ((self.ADDRESS_PATTERN, self.getAddress), (self.DATA_PATTERN, self.parse_line), (self.ETX_PATTERN, self.nop))
        self.formats = [(0, self.ADDRESS_PATTERN), (1, self.DATA_PATTERN), (2, self.ETX_PATTERN)]
        super(ASCIIHexReader, self).__init__()

    def getAddress(self, line, match):
        self.address = int(match.group('address'), 16)
        self.previous_address = self.address
        return True

    def nop(self, line, match):
        return False

    def parse_line(self, line, match):
        section = Section(self.address, bytearray([int(ch, 16) for ch in filter(lambda x: x, self.SPLITTER.split(line))]))
        self.sections.append(section)
        self.address += len(section)
        return True

    def read(self, fp):
        if PYTHON_VERSION.major == 3:
            lines = fp.read().decode()
        else:
            lines = fp.read()
        self.sections = []
        self.address = 0
        breakRequest = False
        for line in lines.splitlines():
            for pattern, action in self.patterns:
                match = pattern.match(line)
                if match:
                    if not action(line, match):
                        breakRequest = True
                    break
            if breakRequest:
                break
        return Image(join_sections(self.sections))


class ASCIIHexWriter(Writer):

    MAX_ADDRESS_BITS = 16
    previous_address = None

    def __init__(self, address_designator):
        self.separator = ' '
        self.address_designator = address_designator
        super(ASCIIHexWriter, self).__init__()

    def compose_row(self, address, length, row):
        prepend_address =  True if address != self.previous_address else False
        self.previous_address = (address + length)
        if prepend_address:
            line = "{0}\n{1}".format("{0}{1:04X}".format(
                self.address_designator, address), "{0}".format(self.separator).join(["{0:02X}".format(x) for x in row])
            )
        else:
            line = " ".join(["{0:02X}".format(x) for x in row])
        self.row_callout(address, length, row)
        return line

    def row_callout(self, address, length, row):
        pass




import unittest
from objutils.utils import cygpath_to_win

class TestCygpath(unittest.TestCase):

    def test01(self):
        self.assertEqual(cygpath_to_win("/cygdrive/c/projects/foobar/flonz"), r"c:\projects\foobar\flonz")

    def test02(self):
        self.assertEqual(cygpath_to_win(r"c:\projects\foobar\flonz"), r"c:\projects\foobar\flonz")


def main():
    unittest.main()

if __name__ == '__main__':
    main()




#!/usr/bin/env python
# -*- coding: utf-8 -*-


import unittest
from objutils.utils import create_string_buffer
from objutils.readers import PlainBinaryReader


class TestReader(unittest.TestCase):

    @staticmethod
    def _createReader():
        return PlainBinaryReader(create_string_buffer(bytearray("hello world!", "ascii")), PlainBinaryReader.LITTLE_ENDIAN)

    def testSize(self):
        reader = self._createReader()
        self.assertEqual(reader.size, 12)

    def testInitialPosition(self):
        reader = self._createReader()
        self.assertEqual(reader.pos, 0)

    def testSetPosition(self):
        reader = self._createReader()
        reader.pos = 6
        self.assertEqual(reader.pos, 6)

    def testResetPosition(self):
        reader = self._createReader()
        reader.pos = 6
        reader.reset()
        self.assertEqual(reader.pos, 0)

    def testNextByte(self):
        reader = self._createReader()
        self.assertEqual(reader.next_byte(), 104)


class Decoder(unittest.TestCase):

    def _runTest(self, method, value, expected):
        dr = PlainBinaryReader(create_string_buffer(value), PlainBinaryReader.LITTLE_ENDIAN)
        self.assertEqual(getattr(dr, method)(), expected)

    def sleb(self, value, expected):
        self._runTest("sleb", bytearray(value), expected)

    def u8(self, value, expected):
        self._runTest("u8", bytearray(value), expected)

    def s8(self, value, expected):
        self._runTest("s8", bytearray(value), expected)

    def u16(self, value, expected):
        self._runTest("u16", bytearray(value), expected)

    def s16(self, value, expected):
        self._runTest("s16", bytearray(value), expected)

    def u32(self, value, expected):
        self._runTest("u32", bytearray(value), expected)

    def s32(self, value, expected):
        self._runTest("s32", bytearray(value), expected)

    def u64(self, value, expected):
        self._runTest("u64", bytearray(value), expected)

    def s64(self, value, expected):
        self._runTest("s64", bytearray(value), expected)


class TestDecoding(Decoder):

    def testSlb01(self):
        self.sleb(b"\x9b\xf1\x59", -624485)

    def testASCIIZValue(self):
        self._runTest("asciiz", b"hello world!\x00", "hello world!")

    def testASCIIZEmpty(self):
        self._runTest("asciiz", b"\x00", "")

    def testU8_1(self):
        self.u8(b"\x00", 0)

    def testU8_2(self):
        self.u8(b"\x80", 128)

    def testU8_3(self):
        self.u8(b"\xff", 255)

    def testS8_1(self):
        self.s8(b"\x00", 0)

    def testS8_2(self):
        self.s8(b"\x80", -128)

    def testS8_3(self):
        self.s8(b"\xff", -1)

    def testU16_1(self):
        self.u16(b"\x00\x00", 0)

    def testU16_2(self):
        self.u16(b"\x80\x00", 128)

    def testU16_3(self):
        self.u16(b"\x00\xff", 65280)

    def testS16_1(self):
        self.s16(b"\x00\x00", 0)

    def testS16_2(self):
        self.s16(b"\x00\x80", -32768)

    def testS16_3(self):
        self.s16(b"\xff\xff", -1)

    def testU32_1(self):
        self.u32(b"\x00\x00\x00\x00", 0)

    def testU32_2(self):
        self.u32(b"\x80\x00\x00\x00", 128)

    def testU32_3(self):
        self.u32(b"\x00\x00\x00\xff", 4278190080)

    def testS32_1(self):
        self.s32(b"\x00\x00\x00\x00", 0)

    def testS32_2(self):
        self.s32(b"\x00\x00\x00\x80", -2147483648)

    def testS32_3(self):
        self.s32(b"\xff\xff\xff\xff", -1)

    def testU64_1(self):
        self.u64(b"\x00\x00\x00\x00\x00\x00\x00\x00", 0)

    def testU64_2(self):
        self.u64(b"\x80\x00\x00\x00\x00\x00\x00\x00", 128)

    def testU64_3(self):
        self.u64(b"\x00\x00\x00\x00\x00\x00\x00\xff", 18374686479671623680)

    def testS64_1(self):
        self.s64(b"\x00\x00\x00\x00\x00\x00\x00\x00", 0)

    def testS64_2(self):
        self.s64(b"\x00\x00\x00\x00\x00\x00\x00\x80", -9223372036854775808)

    def testS64_3(self):
        self.s64(b"\xff\xff\xff\xff\xff\xff\xff\xff", -1)


def main():
    unittest.main()

if __name__ == '__main__':
    main()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest

from objutils import loads
from objutils.image import Image
from objutils import hexfile


class TestHexfile(unittest.TestCase):

    def setUp(self):
        self.image = Image()

    def tearDown(self):
        del self.image


    def testAddSectionAliasWorks(self):
        self.image.insert_section(range(64), 0x1000)
        # Ok, if no exception gets raised.

    def testRaisesInvalidChecksumError(self):
        self.assertRaises(hexfile.InvalidRecordChecksumError, loads, "srec", b'S110000048656C6C6F2C20776F726C6421AA')

    def testRaisesError(self):
#        self.assertRaises(hexfile.InvalidRecordChecksumError, loads, "srec", b'S110000048656C6C6F20776F726C642166')
        pass


def main():
    unittest.main()

if __name__ == '__main__':
    main()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
"""

__version__ = "0.1.0"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2020 by Christoph Schueler <github.com/Christoph2,
                                        cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from array import array
from collections import namedtuple
from copy import copy
from operator import attrgetter
import re
import reprlib
import struct
import sys

import attr
from sortedcontainers import SortedList

from objutils.exceptions import InvalidAddressError, FeatureNotAvailableError
import objutils.hexdump as hexdump
from objutils.utils import PYTHON_VERSION


##
## todo: find/search methode(n) mit slice funktion!
## Basic patch-interface: (addr (plus ext!), datatype (endianess)) - readAs()/writeAs()
## API to change start_address (but not length!) /w check for overlapps.
## split (for some reason) contingous regions into separate segments (splitAt [addresses], splitInto [n pieces]) inplace or new object.
## cut/copy/paste/delete
##

FORMATS = {
    "uint8": "B",
    "int8": "b",
    "uint16": "H",
    "int16": "h",
    "uint32": "I",
    "int32": "i",
    "uint64": "Q",
    "int64": "q",
    "float32": "f",
    "float64": "d",
}

TYPE_SIZES = {
    "uint8":    1,
    "int8":     1,
    "uint16":   2,
    "int16":    2,
    "uint32":   4,
    "int32":    4,
    "uint64":   8,
    "int64":    8,
    "float32":  4,
    "float64":  8,
}


BYTEORDER = {
    "le": "<",
    "be": ">",
}

TypeInformation = namedtuple("TypeInformation", "type byte_order size")

DTYPE = re.compile(r"""
      (?:(?P<uns>u)?int(?P<len>8 | 16 | 32 | 64)(?P<sep>[-/_:])(?P<end> be | le))
    | (?P<byte>byte)
    | (?P<flt>float)(?P<flen>32 | 64)""", re.IGNORECASE | re.VERBOSE)

NumberRange = namedtuple("NumberRange", "lower upper")

signed_range = lambda x: NumberRange( *(int(-(2 ** x / 2)), int((2 ** x / 2) - 1)))
unsigned_range = lambda x: NumberRange( *(0, int((2 ** x) - 1)))

INT8_RANGE  = signed_range(8)
INT16_RANGE = signed_range(16)
INT32_RANGE = signed_range(32)
INT64_RANGE = signed_range(64)
UINT8_RANGE  = unsigned_range(8)
UINT16_RANGE = unsigned_range(16)
UINT32_RANGE = unsigned_range(32)
UINT64_RANGE = unsigned_range(64)

RANGES = {
    "byte": UINT8_RANGE,
    "uint8": UINT8_RANGE,
    "int8": INT8_RANGE,
    "uint16": UINT16_RANGE,
    "int16": INT16_RANGE,
    "uint32": UINT32_RANGE,
    "int32": INT32_RANGE,
    "uint64": UINT64_RANGE,
    "int64": INT64_RANGE,
}

def filler(ch, n):
    """Create an bytearray consisting of `n` `ch`s.

    Parameters
    ----------
    ch: int (0 <= ch <= 255)
        fill char
    n: int
        repetition count

    Returns
    -------
    bytearray
    """
    if not isinstance(ch, int):
        raise TypeError("ch must be of type int")
    if not (0 <= ch < 256):
        raise ValueError("ch must be between 0 and 255")
    if not isinstance(n, int):
        raise TypeError("n must be of type int")
    if n < 1:
        raise ValueError("n must be >= 1")
    return bytearray([ch] * n)


def _data_converter(data):
    if isinstance(data, bytearray):
        pass    # no conversion needed.
    elif isinstance(data, int):
        raise ValueError("single int not permitted")
    elif isinstance(data, str):
        if PYTHON_VERSION.major == 3 :
            data = bytearray(data, encoding = "ascii")
        else:
            data = bytearray(data)
    elif isinstance(data, array) and data.typecode != 'B':
        if PYTHON_VERSION.major == 3:
            data = bytearray(data.tobytes())
        else:
            data = bytearray(data.tostring())
    elif isinstance(data, Section):
        data = copy(data.data)    # just copy data from other section.
    else:
        try:
            data = bytearray(data)
        except Exception:
            raise ValueError("cannot convert '{}' to bytearray()".format(data))
    return data


@attr.s(repr = False, eq = True, order = True)
class Section(object):
    """Manage sections.

    A section is a continuous block of bytes, with a start-address and known length.

    """
    start_address = attr.ib(type = int, eq = True, order = True, default = 0)
    data = attr.ib(default = bytearray(), converter = _data_converter, eq = True, order = True)

    def __attrs_post_init__(self):
        self.repr = reprlib.Repr()
        self.repr.maxstring = 64
        self.repr.maxother = 64

    def __iter__(self):
        yield self

    def hexdump(self, fp = sys.stdout):
        dumper = hexdump.CanonicalDumper(fp)
        dumper.dump_data(self)

    def tobytes(self):
        if PYTHON_VERSION.major == 3:
            return  array('B', self.data).tobytes()
        else:
            return  array('B', self.data).tostring()

    def tolist(self):
        return  array('B', self.data).tolist()

    def _verify_dtype(self, dtype):
        """
        """
        dtype = dtype.lower().strip()
        if dtype == "byte":
            return "uint8", "le"    # Completly arbitrary,
        if not "_" in dtype or not (dtype.endswith("_le") or dtype.endswith("_be")):
                print("DTYPE:", dtype)
                raise TypeError("dtype must be suffixed with '_be' or '_le'")
        match = DTYPE.match(dtype)
        if not match:
            raise TypeError("Invalid datatype '{}'".format(dtype))
        return dtype.split("_")

    def _getformat(self, dtype, length = 1):
        """
        """
        fmt, bo = self._verify_dtype(dtype)

        if length > 1:
            return "{}{}{}".format(BYTEORDER.get(bo), length, FORMATS.get(fmt))
        else:
            return "{}{}".format(BYTEORDER.get(bo), FORMATS.get(fmt))

    def read(self, addr, length, **kws):
        """
        Parameters
        ----------
        addr: int

        length: int
        """
        offset = addr - self.start_address
        if offset < 0:
            raise InvalidAddressError("read() access out of bounds.")
        if offset + length > self.length:
            raise InvalidAddressError("read() access out of bounds.")
        data = self.data[offset : offset + length]
        return data

    def write(self, addr, data, **kws):
        """
        Parameters
        ----------
        addr: int

        length: int

        data: array-like
        """
        length = len(data)
        offset = addr - self.start_address
        if offset < 0:
            raise InvalidAddressError("write() access out of bounds.")
        if offset + length > self.length:
            raise InvalidAddressError("write() access out of bounds.")
        self.data[offset : offset + length] = data

    def read_numeric(self, addr, dtype, **kws):
        offset = addr - self.start_address
        if offset < 0:
            raise InvalidAddressError("read_numeric() access out of bounds.")
        fmt = self._getformat(dtype)
        data_size = struct.calcsize(fmt)
        if offset + data_size > self.length:
            raise InvalidAddressError("read_numeric() access out of bounds.")
        data = self.data[offset : offset + data_size]
        if 'bit_mask' in kws:
            bit_mask = kws.pop("bit_mask")
            data = self.apply_bitmask(data, dtype, bit_mask)
        return struct.unpack(fmt, data)[0]

    def apply_bitmask(self, data, dtype, bit_mask):
        """
        """
        dtype, byteorder = dtype.lower().strip().split("_")
        byteorder = "little" if byteorder == "le" else "big"
        type_size = TYPE_SIZES.get(dtype)
        data_size = len(data)
        tmp = int.from_bytes(data, byteorder, signed = False)
        tmp &= bit_mask
        return tmp.to_bytes(data_size, byteorder, signed = False)

    def write_numeric(self, addr, value, dtype, **kws):
        offset = addr - self.start_address
        if offset < 0:
            raise InvalidAddressError("write_numeric() access out of bounds.")
        fmt = self._getformat(dtype)
        data_size = struct.calcsize(fmt)
        if offset + data_size > self.length:
            raise InvalidAddressError("write_numeric() access out of bounds.")
        if 'bit_mask' in kws:
            bit_mask = kws.pop("bit_mask")
        self.data[offset : offset + data_size] = struct.pack(fmt, value)

    def read_numeric_array(self, addr, length, dtype, **kws):
        offset = addr - self.start_address
        if offset < 0:
            raise InvalidAddressError("read_numeric_array() access out of bounds.")
        fmt = self._getformat(dtype, length)
        data_size = struct.calcsize(fmt)
        if offset + data_size > self.length:
            raise InvalidAddressError("read_numeric_array() access out of bounds.")
        data = self.data[offset : offset + data_size]
        return struct.unpack(fmt, data)

    def write_numeric_array(self, addr, data, dtype, **kws):
        if not hasattr(data, '__iter__'):
            raise TypeError("data must be iterable")
        length = len(data)
        offset = addr - self.start_address
        if offset < 0:
            raise InvalidAddressError("write_numeric_array() access out of bounds.")
        fmt = self._getformat(dtype, length)
        data_size = struct.calcsize(fmt)
        if offset + data_size > self.length:
            raise InvalidAddressError("write_numeric_array() access out of bounds.")
        self.data[offset : offset + data_size] = struct.pack(fmt, *data)

    def read_string(self, addr, encoding = "latin1", length = -1, **kws):
        offset = addr - self.start_address
        if offset < 0:
            raise InvalidAddressError("read_string() access out of bounds.")
        if length == -1:
            pos = self.data[offset : ].find(b'\x00')
        else:
            pos = length
        if pos == -1:
            raise TypeError("Unterminated String!!!")   # TODO: Testcase.
        return self.data[offset : offset + pos].decode(encoding = encoding)

    def write_string(self, addr, value, encoding = "latin1", **kws):
        offset = addr - self.start_address
        if offset < 0:
            raise InvalidAddressError("write_string() access out of bounds.")
        if PYTHON_VERSION.major == 3:
            self.data[offset : offset +  len(value)] = bytes(value, encoding = encoding)
        else:
            self.data[offset : offset +  len(value)] = bytes(value)
        self.data[offset +  len(value)] = 0

    def write_ndarray(self, addr, array, order = None, **kws):
        """

        """
        try:
            import numpy as np
        except ImportError:
            raise FeatureNotAvailableError("write_ndarray() requires Numpy.")
        offset = addr - self.start_address
        if offset < 0:
            raise InvalidAddressError("write_ndarray() access out of bounds.")
        if not isinstance(array, np.ndarray):
            raise TypeError("array must be of type numpy.ndarray.")
        data_size = array.nbytes
        if offset + data_size > self.length:
            raise InvalidAddressError("write_ndarray() access out of bounds.")
        self.data[offset : offset + data_size] = array.tobytes()

    def read_ndarray(self, addr, length, dtype, shape = None, order = None, **kws):
        """

        """
        try:
            import numpy as np
        except ImportError:
            raise FeatureNotAvailableError("read_ndarray() requires Numpy.")
        offset = addr - self.start_address
        if offset < 0:
            raise InvalidAddressError("read_ndarray() access out of bounds.")
        if offset + length > self.length:
            raise InvalidAddressError("read_ndarray() access out of bounds.")

        """
        If the buffer has data that is not in machine byte-order, this should
        be specified as part of the data-type, e.g.::

          >>> dt = np.dtype(int)
          >>> dt = dt.newbyteorder('>')
          >>> np.frombuffer(buf, dtype=dt)
        """
        type_, byte_order = self._verify_dtype(dtype)
        dt = np.dtype(type_)
        dt = dt.newbyteorder(BYTEORDER.get(byte_order))
        arr = np.frombuffer(self.data[offset : offset + length], dtype = dt).reshape(shape)
        if order == "F":
            return arr.T    # Fortran deposit, i.e. col-maj means transposition.
        else:
            return arr


    """
    def write_timestamp():
        pass

    def read_timestamp():
        pass
    """

    def find(self, expr, addr = -1):
        for item in re.finditer(bytes(expr), self.data):
            yield (self.start_address + item.start(), item.end()- item.start())

    def __repr__(self):
        return "Section(address = 0X{0:08X}, length = {1:d}, data = {2})".format(
            self.start_address,
            self.length,
            self.repr.repr(memoryview(self.data).tobytes())
        )

    def __len__(self):
        return len(self.data)

    @property
    def length(self):
        return len(self)

    def __contains__(self, addr):
        return self.start_address <= addr < (self.start_address + self.length)

    @property
    def address(self):  # Alias
        return self.start_address


def join_sections(sections):
    if isinstance(sections, SortedList):
        result_sections = SortedList(key = attrgetter("start_address"))
    else:
        result_sections = []
    prev_section = Section()
    while sections:
        section = sections.pop(0)
        if not isinstance(section, Section):
            raise TypeError("'{}' is not a 'Section' instance", section)
        if section.start_address == prev_section.start_address + prev_section.length and result_sections:
            last_segment = result_sections[-1]
            last_segment.data.extend(section.data)
        else:
            # Create a new section.
            if isinstance(sections, SortedList):
                result_sections.add(Section(section.start_address, section.data))
            else:
                result_sections.append(Section(section.start_address, section.data))
        prev_section = section
    if result_sections:
        return result_sections
    else:
        if isinstance(sections, SortedList):
            return SortedList(key = attrgetter("start_address"))
        else:
            return []



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.1.0"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2019 by Christoph Schueler <github.com/Christoph2,
                                        cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


import abc

DUMMY_PROTOCOL  = None

class PickleIF(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def dump(self, obj, file_, protocol = DUMMY_PROTOCOL): pass

    @abc.abstractmethod
    def dumps(self, obj, protocol = DUMMY_PROTOCOL): pass

    @abc.abstractmethod
    def load(self, file_): pass

    @abc.abstractmethod
    def loads(self, string_): pass




#!/bin/env python

import os
from setuptools import setup, find_packages
from glob import glob
import sys

with open(os.path.join('objutils', 'version.py'), 'r') as f:
    for line in f:
        if line.startswith('__version__'):
            version = line.split('=')[-1].strip().strip('"')
            break

install_reqs = ['future', 'mako', 'six', 'construct', 'attrs >= 19.3.0', 'sortedcontainers', 'SQLAlchemy']

if sys.version_info.major == 2 or (sys.version_info.major == 3 and sys.version_info.minor < 4):
    install_reqs.extend(['enum34', 'mock'])

with open("docs/README.rst", "r") as fh:
    long_description = fh.read()

setup(
    name = 'objutils',
    version = version,
    provides = ['objutils'],
    description = "Objectfile library for Python",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    author = 'Christoph Schueler',
    author_email = 'cpu12.gems@googlemail.com',
    url = 'http://github.com/christoph2/objutils',
    packages=find_packages(),
    #include_package_data=True,
    extras_require={
       "docs": [
           'sphinxcontrib-napoleon'
       ],
        "develop": [
            "bumpversion"
       ]
    } ,
    install_requires = install_reqs,
    entry_points = {
        'console_scripts': [
                'oj-elf-info = objutils.scripts.oj_elf_info:main',
                'oj-elf-syms = objutils.scripts.oj_elf_syms:main',
                'oj-hex-info = objutils.scripts.oj_hex_info:main',
                'oj-elf-arm-attrs = objutils.scripts.oj_elf_arm_attrs:main',
                'oj-elf-extract = objutils.scripts.oj_elf_extract:main',
    #            'ticoff-dump = objutils.tools.ticoffdump:main'
        ],
    },
    tests_require=["pytest", "pytest-runner"],
    test_suite="objutils.tests",
    #data_files = [
    #    ('objutils/tests/ELFFiles', glob('objutils/tests/ELFFiles*.*')),
    #],
    package_dir = {'tests': 'objutils/tests'},
    license='GPLv2',
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    keywords = [
        'hex files',
        'intel hex',
        's19',
        'srec',
        'srecords',
        'object files',
        'map files',
        'embedded',
        'microcontroller',
        'ECU',
        'shf',
        'rfc4194',
    ],
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    #package_data = {'tests': ['ELFFiles/*.*']},
)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.1.0"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2020 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from objutils import hexfile
from objutils import utils
import objutils.checksums as checksums

DATA = 1
EOF = 2


class Reader(hexfile.Reader):

    FORMAT_SPEC = (
        (DATA, ":AAAALLBBDDCC"),
        (EOF, ":00")
    )

    def check_line(self, line, format_type):
        if format_type == DATA:
            if line.length != len(line.chunk):
                raise hexfile.InvalidRecordLengthError("Byte count doesn't match length of actual data.")
            address_checksum = checksums.rotatedXOR(utils.make_list(utils.int_to_array(line.address), line.length), 8, checksums.ROTATE_LEFT)
            if line.addrChecksum != address_checksum:
                raise hexfile.InvalidRecordChecksumError()
            data_checksum = checksums.rotatedXOR(line.chunk, 8, checksums.ROTATE_LEFT)
            if line.checksum != data_checksum:
                raise hexfile.InvalidRecordChecksumError()

    def is_data_line(self,line,format_type):
        return format_type == DATA

class Writer(hexfile.Writer):

    MAX_ADDRESS_BITS = 16

    def compose_row(self, address, length, row):
        address_checksum = checksums.rotatedXOR(utils.make_list(utils.int_to_array(address), length), 8, checksums.ROTATE_LEFT)
        data_checksum = checksums.rotatedXOR(row, 8, checksums.ROTATE_LEFT)
        line = ":{0:04X}{1:02X}{2:02X}{3}{4:02X}".format(address, length, address_checksum, Writer.hex_bytes(row), data_checksum)
        self.last_address = address + length
        return line

    def compose_footer(self, meta):
      return ":{0:04X}00".format(self.last_address)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.1.0"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2020 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import os
import sys
import threading

import six


def ffs(v: int) -> int:
    """Find first set bit."""
    if v == 0:
        return 0
    res = 0
    while (not (v & 1)):
        v >>= 1
        res += 1
    return res

def slicer(iterable, sliceLength, converter = None):
    if converter is None:
        converter = type(iterable)
    length = len(iterable)
    return [converter((iterable[item : item + sliceLength])) for item in range(0, length, sliceLength)]


def make_list(*args):
    result = []
    for arg in args:
        if hasattr(arg, '__iter__'):
            result.extend(list(arg))
        else:
            result.append(arg)
    return result


def int_to_array(value):
    result = []
    while value:
        result.append(value & 0xff)
        value >>= 8
    if result:
        return list(reversed(result))
    else:
        return [0]


class Curry:
    def __init__(self, fun, *args, **kwargs):
        self.fun = fun
        self.pending = args[:]
        self.kwargs = kwargs.copy()

    def __call__(self, *args, **kwargs):
        if kwargs and self.kwargs:
            kw = self.kwargs.copy()
            kw.update(kwargs)
        else:
            kw = kwargs or self.kwargs
        return self.fun(*(self.pending + args), **kw)


identity = lambda self,x: x

def get_python_version():
    return sys.version_info

PYTHON_VERSION = get_python_version()

if PYTHON_VERSION.major == 3:
    from io import BytesIO as StringIO
else:
    try:
        from cStringIO import StringIO
    except ImportError:
        from StringIO import StringIO
        #from io import TextIOWrapper as StringIO
        #io.TextIOWrapper(create_string_buffer())


def create_string_buffer(*args):
    """Create a string with file-like behaviour (StringIO on Python 2.x).
    """
    buf = StringIO(*args)
    return buf

def bin_extractor(fname, offset, length):
    """Extract a junk of data from a file.
    """
    fp = open(fname)
    fp.seek(offset)
    data = fp.read(length)
    return data

CYG_PREFIX = "/cygdrive/"

def cygpath_to_win(path):
    if path.startswith(CYG_PREFIX):
        path = path[len(CYG_PREFIX) : ]
        drive_letter = "{0}:\\".format(path[0])
        path = path[2 : ].replace("/", "\\")
        path = "{0}{1}".format(drive_letter, path)
    return path


import ctypes

class StructureWithEnums(ctypes.Structure):
    """Add missing enum feature to ctypes Structures.
    """
    _map = {}

    def __getattribute__(self, name):
        _map = ctypes.Structure.__getattribute__(self, '_map')
        value = ctypes.Structure.__getattribute__(self, name)
        if name in _map:
            EnumClass = _map[name]
            if isinstance(value, ctypes.Array):
                return [EnumClass(x) for x in value]
            else:
                return EnumClass(value)
        else:
            return value

    def __str__(self):
        result = []
        result.append("struct {0} {{".format(self.__class__.__name__))
        for field in self._fields_:
            attr, attrType = field
            if attr in self._map:
                attrType = self._map[attr]
            value = getattr(self, attr)
            result.append("    {0} [{1}] = {2!r};".format(attr, attrType.__name__, value))
        result.append("};")
        return '\n'.join(result)

    __repr__ = __str__


import subprocess

class CommandError(Exception):
    pass

def runCommand(cmd):
    proc = subprocess.Popen(cmd, shell = True, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
    result = proc.communicate()
    proc.wait()
    if proc.returncode:
        raise CommandError("{0}".format(result[1]))
    return result[0]


class SingletonBase(object):
    _lock = threading.Lock()

    def __new__(cls, *args, **kws):
        # Double-Checked Locking
        if not hasattr(cls, '_instance'):
            try:
                cls._lock.acquire()
                if not hasattr(cls, '_instance'):
                    cls._instance = super(SingletonBase, cls).__new__(cls)
            finally:
                cls._lock.release()
        return cls._instance


"""
class RepresentationMixIn(object):

    def __repr__(self):
        keys = [k for k in self.__dict__ if not (k.startswith('__') and k.endswith('__'))]
        result = []
        result.append("{0!s} {{".format(self.__class__.__name__))
        for key in keys:
            value = getattr(self, key)
            if isinstance(value, int):
                line = "    {0!s} = 0x{1:X}".format(key, value)
            elif isinstance(value, (float, type(None))):
                line = "    {0!s} = {1!s}".format(key, value)
            elif isinstance(value, array):
                line = "    {0!s} = {1!s}".format(key, helper.hexDump(value))
            else:
                line = "    {0!s} = '{1!s}'".format(key, value)
            result.append(line)
        result.append("}")
        return '\n'.join(result)
"""

import mmap

def create_memorymapped_fileview(filename, writeable = False):
    size = os.path.getsize(filename)
    fd = os.open(filename, os.O_RDWR if writeable else os.O_RDONLY)
    if six.PY3:
        return memoryview(mmap.mmap(fd, size, access = mmap.ACCESS_WRITE if writeable else mmap.ACCESS_READ))
    else:
        return mmap.mmap(fd, size, access = mmap.ACCESS_WRITE if writeable else mmap.ACCESS_READ)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.1.0"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2020 by Christoph Schueler <github.com/Christoph2,
                                        cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import struct

from construct.core import Construct, ConstructError, Bytes, CString
from construct.core import singleton
from construct import Int16ul, Int32ul, Int64ul, Int8ul
from construct import Int16ub, Int32ub, Int64ub

import six

class ULEBError(ConstructError): pass
class SLEBError(ConstructError): pass


@singleton
class ULEB(Construct):

    def __init__(self, *args):
        if six.PY3:
            super(__class__, self).__init__()
        else:
            super(self.__class__, self).__init__()

    def _parse(self, stream, context, path = None):
        result = 0
        shift = 0
        while True:
            try:
                bval = ord(stream.read(1))
            except Exception as e:
                raise ULEBError(str(e))
            result |= ((bval & 0x7f) << shift)
            if bval & 0x80 == 0:
                break
            shift += 7
        return result


    def _build(self, value, stream, context, path):
        assert value is not None
        if value < 0:
            raise ULEBError('value must be non-negative.')
        result = []
        while True:
            bval = value & 0x7f
            value >>= 7
            if value != 0:
                bval |= 0x80
            result.append(bval)
            if value == 0:
                break
        stream.write(bytes(result))


@singleton
class SLEB(Construct):

    def __init__(self, *args):
        if six.PY3:
            super(__class__, self).__init__()
        else:
            super().__init__()

    def _parse(self, stream, context, path = None):
        result = 0
        shift = 0
        size = 32
        idx = 0
        while True:
            try:
                bval = ord(stream.read(1))
            except Exception as e:
                raise SLEBError(str(e))
            result |= ((bval & 0x7f) << shift)
            shift += 7
            idx += 1
            if bval & 0x80 == 0:
                break
        if (shift < 32) or (bval & 0x40) == 0x40:
            mask = - (1 << (idx * 7))
            result |= mask
        return result

    def _build(self, value, stream, context, path):
        assert value is not None
        result = []
        more = 1
        size = 32
        while more:
            bval = value & 0x7f
            value >>= 7
            if ((value == 0 and (bval & 0x40) == 0x00)) or ((value == -1 and (bval & 0x40) == 0x40)):
                more = 0
            else:
                bval |= 0x80
            result.append(bval)
        stream.write(bytes(result))

@singleton
class One(Construct):

    def __init__(self, *args):
        if six.PY3:
            super(__class__, self).__init__()
        else:
            super().__init__()

    def _parse(self, stream, context, path = None):
        return 1

    def _build(self, value, stream, context, path):
        count = struct.pack("B", 1)
        value = 1
        stream.write(count)
        if six.PY3:
            stream.write(bytes(value, encoding = "ascii"))
        else:
            stream.write(bytes(value))

class Block(Construct):

    BYTEORDER = ''
    SIZE = None
    MASK = None

    def __init__(self, *args):
        if six.PY3:
            super(__class__, self).__init__()
        else:
            super().__init__()

    def _parse(self, stream, context, path = None):
        msk = "{}{}".format(self.BYTEORDER, self.MASK)
        count = stream.read(self.SIZE)
        count = struct.unpack(msk, count)[0]
        result = Bytes(count).parse_stream(stream)
        return result

    def _build(self, value, stream, context, path):
        msk = "{}{}".format(self.BYTEORDER, self.MASK)
        count = struct.pack(msk, len(value))
        stream.write(count)
        if six.PY3:
            stream.write(bytes(value, encoding = "ascii"))
        else:
            stream.write(bytes(value))


@singleton
class BlockUleb(Block):

    def _parse(self, stream, context, path = None):
        count = ULEB.parse_stream(stream)
        result = Bytes(count).parse_stream(stream)
        return result

    def _build(self, value, stream, context, path):
        count = ULEB.build(len(value))
        stream.write(count)
        if six.PY3:
            stream.write(bytes(value, encoding = "ascii"))
        else:
            stream.write(bytes(value))

@singleton
class Block1(Block):

    SIZE = 1
    MASK = "B"
    BYTEORDER = "<"


@singleton
class Block2l(Block):

    SIZE = 2
    MASK = "H"
    BYTEORDER = "<"


@singleton
class Block2b(Block):

    SIZE = 2
    MASK = "H"
    BYTEORDER = ">"


@singleton
class Block4l(Block):

    SIZE = 4
    MASK = "L"
    BYTEORDER = "<"


@singleton
class Block4b(Block):

    SIZE = 4
    MASK = "L"
    BYTEORDER = ">"


class Address(Construct):

    TYPES = {
        1: (Int8ul, Int8ul),
        2: (Int16ul, Int16ub),
        4: (Int32ul, Int32ub),
        8: (Int64ul, Int64ub),
    }

    def __init__(self, size, endianess):
        if six.PY3:
            super(__class__, self).__init__()
        else:
            super().__init__()
        idx = endianess
        if size not in (1, 2, 4, 8):
            raise ValueError("Address size '{}' not supported.".format(size))
        if not isinstance(size, int):
            print("**SIZE is not int", size, endianess, idx, type(str))
        if not isinstance(idx, int):
            print("**IDX is not int", size, endianess, idx, type(idx))
        self.type = self.TYPES[size][idx]

    def _parse(self, stream, context, path = None):
        return self.type.parse_stream(stream)

    def _build(self, value, stream, context, path):
        stream.write(self.type.build(value))


class StrP(Construct):

    BYTEORDER = ''
    SIZE = None
    MASK = None

    def __init__(self, image, endianess):
        self.image = image
        self.endianess = endianess
        self.ntype = (Int32ul, Int32ub)[0 if endianess == 1 else 1]
        self.stype = CString(encoding = "utf8")
        if six.PY3:
            super(__class__, self).__init__()
        else:
            super().__init__()

    def _parse(self, stream, context, path = None):
        offset = self.ntype.parse_stream(stream)
        #self.image.seek(offset)
        data = self.image[offset : ]
        result = self.stype.parse(data)
        return result

    def _build(self, value, stream, context, path):
        stream.write(self.type.build(value))
